from __future__ import annotations

import atexit
import os
import platform
import threading
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import ExitStack
from contextvars import copy_context
from datetime import datetime, timezone
from logging import Logger, getLogger
from types import TracebackType
from typing import Callable
from uuid import UUID

import attrs

from .. import JobReleased
from .._context import current_job, current_worker
from .._enums import JobOutcome, RunState
from .._events import JobAdded, WorkerStarted, WorkerStopped
from .._structures import Job, JobInfo, JobResult
from .._validators import positive_integer
from ..abc import DataStore, EventBroker
from ..eventbrokers.local import LocalEventBroker


@attrs.define(eq=False)
class Worker:
    """Runs jobs locally in a thread pool."""

    data_store: DataStore
    event_broker: EventBroker = attrs.field(factory=LocalEventBroker)
    max_concurrent_jobs: int = attrs.field(
        kw_only=True, validator=positive_integer, default=20
    )
    identity: str = attrs.field(kw_only=True, default=None)
    logger: Logger | None = attrs.field(kw_only=True, default=getLogger(__name__))
    # True if a scheduler owns this worker
    _is_internal: bool = attrs.field(kw_only=True, default=False)

    _state: RunState = attrs.field(init=False, default=RunState.stopped)
    _thread: threading.Thread | None = attrs.field(init=False, default=None)
    _wakeup_event: threading.Event = attrs.field(init=False, factory=threading.Event)
    _executor: ThreadPoolExecutor = attrs.field(init=False)
    _acquired_jobs: set[Job] = attrs.field(init=False, factory=set)
    _running_jobs: set[UUID] = attrs.field(init=False, factory=set)

    def __attrs_post_init__(self) -> None:
        if not self.identity:
            self.identity = f"{platform.node()}-{os.getpid()}-{id(self)}"

    def __enter__(self) -> Worker:
        self.start_in_background()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.stop()

    @property
    def state(self) -> RunState:
        """The current running state of the worker."""
        return self._state

    def start_in_background(self) -> None:
        """
        Launch the worker in a new thread.

        This method registers an :mod:`atexit` hook to shut down the worker and wait
        for the thread to finish.

        """
        start_future: Future[None] = Future()
        self._thread = threading.Thread(
            target=copy_context().run, args=[self._run, start_future], daemon=True
        )
        self._thread.start()
        try:
            start_future.result()
        except BaseException:
            self._thread = None
            raise

        atexit.register(self.stop)

    def stop(self) -> None:
        """
        Signal the worker that it should stop running jobs.

        This method does not wait for the worker to actually stop.

        """
        atexit.unregister(self.stop)
        if self._state is RunState.started:
            self._state = RunState.stopping
            self._wakeup_event.set()

            if threading.current_thread() != self._thread:
                self._thread.join()
                self._thread = None

    def run_until_stopped(self) -> None:
        """
        Run the worker until it is explicitly stopped.

        This method will only return if :meth:`stop` is called.

        """
        self._run(None)

    def _run(self, start_future: Future[None] | None) -> None:
        with ExitStack() as exit_stack:
            try:
                if self._state is not RunState.stopped:
                    raise RuntimeError(
                        f'Cannot start the worker when it is in the "{self._state}" '
                        f"state"
                    )

                if not self._is_internal:
                    # Initialize the event broker
                    self.event_broker.start()
                    exit_stack.push(
                        lambda *exc_info: self.event_broker.stop(
                            force=exc_info[0] is not None
                        )
                    )

                    # Initialize the data store
                    self.data_store.start(self.event_broker)
                    exit_stack.push(
                        lambda *exc_info: self.data_store.stop(
                            force=exc_info[0] is not None
                        )
                    )

                # Set the current worker
                token = current_worker.set(self)
                exit_stack.callback(current_worker.reset, token)

                # Wake up the worker if the data store emits a significant job event
                exit_stack.enter_context(
                    self.event_broker.subscribe(
                        lambda event: self._wakeup_event.set(), {JobAdded}
                    )
                )

                # Initialize the thread pool
                executor = ThreadPoolExecutor(max_workers=self.max_concurrent_jobs)
                exit_stack.enter_context(executor)

                # Signal that the worker has started
                self._state = RunState.started
                self.event_broker.publish_local(WorkerStarted())
            except BaseException as exc:
                if start_future:
                    start_future.set_exception(exc)
                    return
                else:
                    raise
            else:
                if start_future:
                    start_future.set_result(None)

            try:
                while self._state is RunState.started:
                    available_slots = self.max_concurrent_jobs - len(self._running_jobs)
                    if available_slots:
                        jobs = self.data_store.acquire_jobs(
                            self.identity, available_slots
                        )
                        for job in jobs:
                            task = self.data_store.get_task(job.task_id)
                            self._running_jobs.add(job.id)
                            executor.submit(
                                copy_context().run, self._run_job, job, task.func
                            )

                    self._wakeup_event.wait()
                    self._wakeup_event = threading.Event()
            except BaseException as exc:
                self._state = RunState.stopped
                if isinstance(exc, Exception):
                    self.logger.exception("Worker crashed")
                else:
                    self.logger.info(f"Worker stopped due to {exc.__class__.__name__}")

                self.event_broker.publish_local(WorkerStopped(exception=exc))
            else:
                self._state = RunState.stopped
                self.logger.info("Worker stopped")
                self.event_broker.publish_local(WorkerStopped())

    def _run_job(self, job: Job, func: Callable) -> None:
        try:
            # Check if the job started before the deadline
            start_time = datetime.now(timezone.utc)
            if job.start_deadline is not None and start_time > job.start_deadline:
                result = JobResult.from_job(
                    job, JobOutcome.missed_start_deadline, finished_at=start_time
                )
                self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
                self.data_store.release_job(self.identity, job.task_id, result)
                return

            token = current_job.set(JobInfo.from_job(job))
            try:
                retval = func(*job.args, **job.kwargs)
            except BaseException as exc:
                if isinstance(exc, Exception):
                    self.logger.exception("Job %s raised an exception", job.id)
                else:
                    self.logger.error(
                        "Job %s was aborted due to %s", job.id, exc.__class__.__name__
                    )

                result = JobResult.from_job(
                    job,
                    JobOutcome.error,
                    exception=exc,
                )
                self.data_store.release_job(
                    self.identity,
                    job.task_id,
                    result,
                )
                self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
                if not isinstance(exc, Exception):
                    raise
            else:
                self.logger.info("Job %s completed successfully", job.id)
                result = JobResult.from_job(
                    job,
                    JobOutcome.success,
                    return_value=retval,
                )
                self.data_store.release_job(self.identity, job.task_id, result)
                self.event_broker.publish(
                    JobReleased.from_result(result, self.identity)
                )
            finally:
                current_job.reset(token)
        finally:
            self._running_jobs.remove(job.id)



from __future__ import annotations

from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._structures import JobInfo
    from .schedulers.async_ import AsyncScheduler
    from .schedulers.sync import Scheduler
    from .workers.async_ import AsyncWorker
    from .workers.sync import Worker

#: The currently running (local) scheduler
current_scheduler: ContextVar[Scheduler | AsyncScheduler | None] = ContextVar(
    "current_scheduler", default=None
)
#: The worker running the current job
current_worker: ContextVar[Worker | AsyncWorker | None] = ContextVar(
    "current_worker", default=None
)
#: Metadata about the current job
current_job: ContextVar[JobInfo] = ContextVar("job_info")



from __future__ import annotations

import threading
from datetime import datetime, timezone
from typing import Callable

import anyio
import pytest
from anyio import fail_after

from apscheduler import (
    Event,
    Job,
    JobAcquired,
    JobAdded,
    JobOutcome,
    JobReleased,
    TaskAdded,
    WorkerStopped,
)
from apscheduler._structures import Task
from apscheduler.datastores.memory import MemoryDataStore
from apscheduler.workers.async_ import AsyncWorker
from apscheduler.workers.sync import Worker

pytestmark = pytest.mark.anyio


def sync_func(*args, fail: bool, **kwargs):
    if fail:
        raise Exception("failing as requested")
    else:
        return args, kwargs


async def async_func(*args, fail: bool, **kwargs):
    if fail:
        raise Exception("failing as requested")
    else:
        return args, kwargs


def fail_func():
    pytest.fail("This function should never be run")


class TestAsyncWorker:
    @pytest.mark.parametrize(
        "target_func", [sync_func, async_func], ids=["sync", "async"]
    )
    @pytest.mark.parametrize("fail", [False, True], ids=["success", "fail"])
    async def test_run_job_nonscheduled_success(
        self, target_func: Callable, fail: bool
    ) -> None:
        def listener(received_event: Event):
            received_events.append(received_event)
            if isinstance(received_event, JobReleased):
                event.set()

        received_events: list[Event] = []
        event = anyio.Event()
        async with AsyncWorker(MemoryDataStore()) as worker:
            worker.event_broker.subscribe(listener)
            await worker.data_store.add_task(Task(id="task_id", func=target_func))
            job = Job(task_id="task_id", args=(1, 2), kwargs={"x": "foo", "fail": fail})
            await worker.data_store.add_job(job)
            with fail_after(3):
                await event.wait()

        # First, a task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "task_id"

        # Then a job was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.job_id == job.id
        assert received_event.task_id == "task_id"
        assert received_event.schedule_id is None

        # Then the job was started
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAcquired)
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        received_event = received_events.pop(0)
        if fail:
            # Then the job failed
            assert isinstance(received_event, JobReleased)
            assert received_event.outcome is JobOutcome.error
            assert received_event.exception_type == "Exception"
            assert received_event.exception_message == "failing as requested"
            assert isinstance(received_event.exception_traceback, list)
            assert all(
                isinstance(line, str) for line in received_event.exception_traceback
            )
        else:
            # Then the job finished successfully
            assert isinstance(received_event, JobReleased)
            assert received_event.outcome is JobOutcome.success
            assert received_event.exception_type is None
            assert received_event.exception_message is None
            assert received_event.exception_traceback is None

        # Finally, the worker was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, WorkerStopped)

        # There should be no more events on the list
        assert not received_events

    async def test_run_deadline_missed(self) -> None:
        def listener(received_event: Event):
            received_events.append(received_event)
            if isinstance(received_event, JobReleased):
                event.set()

        scheduled_start_time = datetime(2020, 9, 14, tzinfo=timezone.utc)
        received_events: list[Event] = []
        event = anyio.Event()
        async with AsyncWorker(MemoryDataStore()) as worker:
            worker.event_broker.subscribe(listener)
            await worker.data_store.add_task(Task(id="task_id", func=fail_func))
            job = Job(
                task_id="task_id",
                schedule_id="foo",
                scheduled_fire_time=scheduled_start_time,
                start_deadline=datetime(2020, 9, 14, 1, tzinfo=timezone.utc),
            )
            await worker.data_store.add_job(job)
            with fail_after(3):
                await event.wait()

        # First, a task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "task_id"

        # Then a job was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.job_id == job.id
        assert received_event.task_id == "task_id"
        assert received_event.schedule_id == "foo"

        # The worker acquired the job
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAcquired)
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        # The worker determined that the deadline has been missed
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobReleased)
        assert received_event.outcome is JobOutcome.missed_start_deadline
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        # Finally, the worker was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, WorkerStopped)

        # There should be no more events on the list
        assert not received_events


class TestSyncWorker:
    @pytest.mark.parametrize("fail", [False, True], ids=["success", "fail"])
    def test_run_job_nonscheduled(self, fail: bool) -> None:
        def listener(received_event: Event):
            received_events.append(received_event)
            if isinstance(received_event, JobReleased):
                event.set()

        received_events: list[Event] = []
        event = threading.Event()
        with Worker(MemoryDataStore()) as worker:
            worker.event_broker.subscribe(listener)
            worker.data_store.add_task(Task(id="task_id", func=sync_func))
            job = Job(task_id="task_id", args=(1, 2), kwargs={"x": "foo", "fail": fail})
            worker.data_store.add_job(job)
            event.wait(3)

        # First, a task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "task_id"

        # Then a job was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.job_id == job.id
        assert received_event.task_id == "task_id"
        assert received_event.schedule_id is None

        # Then the job was started
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAcquired)
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        received_event = received_events.pop(0)
        if fail:
            # Then the job failed
            assert isinstance(received_event, JobReleased)
            assert received_event.outcome is JobOutcome.error
            assert received_event.exception_type == "Exception"
            assert received_event.exception_message == "failing as requested"
            assert isinstance(received_event.exception_traceback, list)
            assert all(
                isinstance(line, str) for line in received_event.exception_traceback
            )
        else:
            # Then the job finished successfully
            assert isinstance(received_event, JobReleased)
            assert received_event.outcome is JobOutcome.success
            assert received_event.exception_type is None
            assert received_event.exception_message is None
            assert received_event.exception_traceback is None

        # Finally, the worker was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, WorkerStopped)

        # There should be no more events on the list
        assert not received_events

    def test_run_deadline_missed(self) -> None:
        def listener(received_event: Event):
            received_events.append(received_event)
            if isinstance(received_event, JobReleased):
                event.set()

        scheduled_start_time = datetime(2020, 9, 14, tzinfo=timezone.utc)
        received_events: list[Event] = []
        event = threading.Event()
        with Worker(MemoryDataStore()) as worker:
            worker.event_broker.subscribe(listener)
            worker.data_store.add_task(Task(id="task_id", func=fail_func))
            job = Job(
                task_id="task_id",
                schedule_id="foo",
                scheduled_fire_time=scheduled_start_time,
                start_deadline=datetime(2020, 9, 14, 1, tzinfo=timezone.utc),
            )
            worker.data_store.add_job(job)
            event.wait(3)

        # First, a task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "task_id"

        # Then a job was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.job_id == job.id
        assert received_event.task_id == "task_id"
        assert received_event.schedule_id == "foo"

        # The worker acquired the job
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAcquired)
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        # The worker determined that the deadline has been missed
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobReleased)
        assert received_event.outcome is JobOutcome.missed_start_deadline
        assert received_event.job_id == job.id
        assert received_event.worker_id == worker.identity

        # Finally, the worker was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, WorkerStopped)

        # There should be no more events on the list
        assert not received_events



"""
Example demonstrating use with the Starlette web framework.

Requires the "postgresql" service to be running.
To install prerequisites: pip install sqlalchemy asycnpg starlette uvicorn
To run: uvicorn asgi_starlette:app

It should print a line on the console on a one-second interval while running a
basic web app at http://localhost:8000.
"""

from __future__ import annotations

from datetime import datetime

from sqlalchemy.ext.asyncio import create_async_engine
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Route
from starlette.types import ASGIApp, Receive, Scope, Send

from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore
from apscheduler.eventbrokers.asyncpg import AsyncpgEventBroker
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


class SchedulerMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        scheduler: AsyncScheduler,
    ) -> None:
        self.app = app
        self.scheduler = scheduler

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":
            async with self.scheduler:
                await self.scheduler.add_schedule(
                    tick, IntervalTrigger(seconds=1), id="tick"
                )
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)


async def root(request: Request) -> Response:
    return PlainTextResponse("Hello, world!")


engine = create_async_engine("postgresql+asyncpg://postgres:secret@localhost/testdb")
data_store = AsyncSQLAlchemyDataStore(engine)
event_broker = AsyncpgEventBroker.from_async_sqla_engine(engine)
scheduler = AsyncScheduler()
routes = [Route("/", root)]
middleware = [Middleware(SchedulerMiddleware, scheduler=scheduler)]
app = Starlette(routes=routes, middleware=middleware)



"""This module contains several handy functions primarily meant for internal use."""
from __future__ import annotations

import sys
from datetime import datetime, tzinfo
from typing import TypeVar

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo

T = TypeVar("T")


def timezone_repr(timezone: tzinfo) -> str:
    if isinstance(timezone, ZoneInfo):
        return timezone.key
    else:
        return repr(timezone)


def absolute_datetime_diff(dateval1: datetime, dateval2: datetime) -> float:
    return dateval1.timestamp() - dateval2.timestamp()


def qualified_name(cls: type) -> str:
    module = getattr(cls, "__module__", None)
    if module is None or module == "builtins":
        return cls.__qualname__
    else:
        return f"{module}.{cls.__qualname__}"



from __future__ import annotations

from typing import Any, Callable, Iterable

import attrs
from anyio import to_thread
from anyio.from_thread import BlockingPortal

from .._events import Event
from ..abc import AsyncEventBroker, EventBroker, Subscription


@attrs.define(eq=False)
class AsyncEventBrokerAdapter(AsyncEventBroker):
    original: EventBroker

    async def start(self) -> None:
        await to_thread.run_sync(self.original.start)

    async def stop(self, *, force: bool = False) -> None:
        await to_thread.run_sync(lambda: self.original.stop(force=force))

    async def publish_local(self, event: Event) -> None:
        await to_thread.run_sync(self.original.publish_local, event)

    async def publish(self, event: Event) -> None:
        await to_thread.run_sync(self.original.publish, event)

    def subscribe(
        self,
        callback: Callable[[Event], Any],
        event_types: Iterable[type[Event]] | None = None,
        *,
        one_shot: bool = False,
    ) -> Subscription:
        return self.original.subscribe(callback, event_types, one_shot=one_shot)


@attrs.define(eq=False)
class SyncEventBrokerAdapter(EventBroker):
    original: AsyncEventBroker
    portal: BlockingPortal

    def start(self) -> None:
        pass

    def stop(self, *, force: bool = False) -> None:
        pass

    def publish_local(self, event: Event) -> None:
        self.portal.call(self.original.publish_local, event)

    def publish(self, event: Event) -> None:
        self.portal.call(self.original.publish, event)

    def subscribe(
        self,
        callback: Callable[[Event], Any],
        event_types: Iterable[type[Event]] | None = None,
        *,
        one_shot: bool = False,
    ) -> Subscription:
        return self.portal.call(
            lambda: self.original.subscribe(callback, event_types, one_shot=one_shot)
        )



"""
Example demonstrating use of the asynchronous scheduler with persistence via MySQL or
MariaDB in a simple asyncio app.

Requires the "mysql" service to be running.
To install prerequisites: pip install sqlalchemy asyncmy
To run: python async_postgres.py

It should print a line on the console on a one-second interval.
"""

from __future__ import annotations

from asyncio import run
from datetime import datetime

from sqlalchemy.ext.asyncio import create_async_engine

from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


async def main():
    engine = create_async_engine(
        "mysql+asyncmy://root:secret@localhost/testdb?charset=utf8mb4"
    )
    data_store = AsyncSQLAlchemyDataStore(engine)
    async with AsyncScheduler(data_store) as scheduler:
        await scheduler.add_schedule(tick, IntervalTrigger(seconds=1), id="tick")
        await scheduler.wait_until_stopped()


run(main())



from __future__ import annotations

import sys

import pytest

from apscheduler.abc import Serializer
from apscheduler.serializers.cbor import CBORSerializer
from apscheduler.serializers.json import JSONSerializer
from apscheduler.serializers.pickle import PickleSerializer

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo


@pytest.fixture(scope="session")
def timezone() -> ZoneInfo:
    return ZoneInfo("Europe/Berlin")


@pytest.fixture(
    params=[
        pytest.param(PickleSerializer, id="pickle"),
        pytest.param(CBORSerializer, id="cbor"),
        pytest.param(JSONSerializer, id="json"),
    ]
)
def serializer(request) -> Serializer | None:
    return request.param() if request.param else None


@pytest.fixture
def anyio_backend() -> str:
    return "asyncio"



from __future__ import annotations

import os
import platform
import random
from contextlib import AsyncExitStack
from datetime import datetime, timedelta, timezone
from logging import Logger, getLogger
from typing import Any, Callable, Iterable, Mapping, cast
from uuid import UUID, uuid4

import anyio
import attrs
from anyio import TASK_STATUS_IGNORED, create_task_group, move_on_after
from anyio.abc import TaskGroup, TaskStatus

from .._context import current_scheduler
from .._converters import as_async_datastore, as_async_eventbroker
from .._enums import CoalescePolicy, ConflictPolicy, JobOutcome, RunState
from .._events import (
    Event,
    JobReleased,
    ScheduleAdded,
    SchedulerStarted,
    SchedulerStopped,
    ScheduleUpdated,
)
from .._exceptions import (
    JobCancelled,
    JobDeadlineMissed,
    JobLookupError,
    ScheduleLookupError,
)
from .._structures import Job, JobResult, Schedule, Task
from ..abc import AsyncDataStore, AsyncEventBroker, Subscription, Trigger
from ..datastores.memory import MemoryDataStore
from ..eventbrokers.async_local import LocalAsyncEventBroker
from ..marshalling import callable_to_ref
from ..workers.async_ import AsyncWorker

_microsecond_delta = timedelta(microseconds=1)
_zero_timedelta = timedelta()


@attrs.define(eq=False)
class AsyncScheduler:
    """An asynchronous (AnyIO based) scheduler implementation."""

    data_store: AsyncDataStore = attrs.field(
        converter=as_async_datastore, factory=MemoryDataStore
    )
    event_broker: AsyncEventBroker = attrs.field(
        converter=as_async_eventbroker, factory=LocalAsyncEventBroker
    )
    identity: str = attrs.field(kw_only=True, default=None)
    start_worker: bool = attrs.field(kw_only=True, default=True)
    logger: Logger | None = attrs.field(kw_only=True, default=getLogger(__name__))

    _state: RunState = attrs.field(init=False, default=RunState.stopped)
    _task_group: TaskGroup | None = attrs.field(init=False, default=None)
    _wakeup_event: anyio.Event = attrs.field(init=False)
    _wakeup_deadline: datetime | None = attrs.field(init=False, default=None)
    _schedule_added_subscription: Subscription = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if not self.identity:
            self.identity = f"{platform.node()}-{os.getpid()}-{id(self)}"

    async def __aenter__(self):
        self._task_group = create_task_group()
        await self._task_group.__aenter__()
        await self._task_group.start(self._run)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
        await self._task_group.__aexit__(exc_type, exc_val, exc_tb)
        self._task_group = None

    def _schedule_added_or_modified(self, event: Event) -> None:
        event_ = cast("ScheduleAdded | ScheduleUpdated", event)
        if not self._wakeup_deadline or (
            event_.next_fire_time and event_.next_fire_time < self._wakeup_deadline
        ):
            self.logger.debug(
                "Detected a %s event – waking up the scheduler", type(event).__name__
            )
            self._wakeup_event.set()

    @property
    def state(self) -> RunState:
        """The current running state of the scheduler."""
        return self._state

    async def add_schedule(
        self,
        func_or_task_id: str | Callable,
        trigger: Trigger,
        *,
        id: str | None = None,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        coalesce: CoalescePolicy = CoalescePolicy.latest,
        misfire_grace_time: float | timedelta | None = None,
        max_jitter: float | timedelta | None = None,
        tags: Iterable[str] | None = None,
        conflict_policy: ConflictPolicy = ConflictPolicy.do_nothing,
    ) -> str:
        """
        Schedule a task to be run one or more times in the future.

        :param func_or_task_id: either a callable or an ID of an existing task
            definition
        :param trigger: determines the times when the task should be run
        :param id: an explicit identifier for the schedule (if omitted, a random, UUID
            based ID will be assigned)
        :param args: positional arguments to be passed to the task function
        :param kwargs: keyword arguments to be passed to the task function
        :param coalesce: determines what to do when processing the schedule if multiple
            fire times have become due for this schedule since the last processing
        :param misfire_grace_time: maximum number of seconds the scheduled job's actual
            run time is allowed to be late, compared to the scheduled run time
        :param max_jitter: maximum number of seconds to randomly add to the scheduled
            time for each job created from this schedule
        :param tags: strings that can be used to categorize and filter the schedule and
            its derivative jobs
        :param conflict_policy: determines what to do if a schedule with the same ID
            already exists in the data store
        :return: the ID of the newly added schedule

        """
        id = id or str(uuid4())
        args = tuple(args or ())
        kwargs = dict(kwargs or {})
        tags = frozenset(tags or ())
        if isinstance(misfire_grace_time, (int, float)):
            misfire_grace_time = timedelta(seconds=misfire_grace_time)

        if callable(func_or_task_id):
            task = Task(id=callable_to_ref(func_or_task_id), func=func_or_task_id)
            await self.data_store.add_task(task)
        else:
            task = await self.data_store.get_task(func_or_task_id)

        schedule = Schedule(
            id=id,
            task_id=task.id,
            trigger=trigger,
            args=args,
            kwargs=kwargs,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            tags=tags,
        )
        schedule.next_fire_time = trigger.next()
        await self.data_store.add_schedule(schedule, conflict_policy)
        self.logger.info(
            "Added new schedule (task=%r, trigger=%r); next run time at %s",
            task,
            trigger,
            schedule.next_fire_time,
        )
        return schedule.id

    async def get_schedule(self, id: str) -> Schedule:
        """
        Retrieve a schedule from the data store.

        :param id: the unique identifier of the schedule
        :raises ScheduleLookupError: if the schedule could not be found

        """
        schedules = await self.data_store.get_schedules({id})
        if schedules:
            return schedules[0]
        else:
            raise ScheduleLookupError(id)

    async def get_schedules(self) -> list[Schedule]:
        """
        Retrieve all schedules from the data store.

        :return: a list of schedules, in an unspecified order

        """
        return await self.data_store.get_schedules()

    async def remove_schedule(self, id: str) -> None:
        """
        Remove the given schedule from the data store.

        :param id: the unique identifier of the schedule

        """
        await self.data_store.remove_schedules({id})

    async def add_job(
        self,
        func_or_task_id: str | Callable,
        *,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        tags: Iterable[str] | None = None,
        result_expiration_time: timedelta | float = 0,
    ) -> UUID:
        """
        Add a job to the data store.

        :param func_or_task_id:
        :param args: positional arguments to call the target callable with
        :param kwargs: keyword arguments to call the target callable with
        :param tags: strings that can be used to categorize and filter the job
        :param result_expiration_time: the minimum time (as seconds, or timedelta) to
            keep the result of the job available for fetching (the result won't be
            saved at all if that time is 0)
        :return: the ID of the newly created job

        """
        if callable(func_or_task_id):
            task = Task(id=callable_to_ref(func_or_task_id), func=func_or_task_id)
            await self.data_store.add_task(task)
        else:
            task = await self.data_store.get_task(func_or_task_id)

        job = Job(
            task_id=task.id,
            args=args or (),
            kwargs=kwargs or {},
            tags=tags or frozenset(),
            result_expiration_time=result_expiration_time,
        )
        await self.data_store.add_job(job)
        return job.id

    async def get_job_result(self, job_id: UUID, *, wait: bool = True) -> JobResult:
        """
        Retrieve the result of a job.

        :param job_id: the ID of the job
        :param wait: if ``True``, wait until the job has ended (one way or another),
            ``False`` to raise an exception if the result is not yet available
        :raises JobLookupError: if ``wait=False`` and the job result does not exist in
            the data store

        """
        wait_event = anyio.Event()

        def listener(event: JobReleased) -> None:
            if event.job_id == job_id:
                wait_event.set()

        with self.data_store.events.subscribe(listener, {JobReleased}):
            result = await self.data_store.get_job_result(job_id)
            if result:
                return result
            elif not wait:
                raise JobLookupError(job_id)

            await wait_event.wait()

        return await self.data_store.get_job_result(job_id)

    async def run_job(
        self,
        func_or_task_id: str | Callable,
        *,
        args: Iterable | None = None,
        kwargs: Mapping[str, Any] | None = None,
        tags: Iterable[str] | None = (),
    ) -> Any:
        """
        Convenience method to add a job and then return its result.

        If the job raised an exception, that exception will be reraised here.

        :param func_or_task_id: either a callable or an ID of an existing task
            definition
        :param args: positional arguments to be passed to the task function
        :param kwargs: keyword arguments to be passed to the task function
        :param tags: strings that can be used to categorize and filter the job
        :returns: the return value of the task function

        """
        job_complete_event = anyio.Event()

        def listener(event: JobReleased) -> None:
            if event.job_id == job_id:
                job_complete_event.set()

        job_id: UUID | None = None
        with self.data_store.events.subscribe(listener, {JobReleased}):
            job_id = await self.add_job(
                func_or_task_id,
                args=args,
                kwargs=kwargs,
                tags=tags,
                result_expiration_time=timedelta(minutes=15),
            )
            await job_complete_event.wait()

        result = await self.get_job_result(job_id)
        if result.outcome is JobOutcome.success:
            return result.return_value
        elif result.outcome is JobOutcome.error:
            raise result.exception
        elif result.outcome is JobOutcome.missed_start_deadline:
            raise JobDeadlineMissed
        elif result.outcome is JobOutcome.cancelled:
            raise JobCancelled
        else:
            raise RuntimeError(f"Unknown job outcome: {result.outcome}")

    async def stop(self) -> None:
        """
        Signal the scheduler that it should stop processing schedules.

        This method does not wait for the scheduler to actually stop.
        For that, see :meth:`wait_until_stopped`.

        """
        if self._state is RunState.started:
            self._state = RunState.stopping
            self._wakeup_event.set()

    async def wait_until_stopped(self) -> None:
        """
        Wait until the scheduler is in the "stopped" or "stopping" state.

        If the scheduler is already stopped or in the process of stopping, this method
        returns immediately. Otherwise, it waits until the scheduler posts the
        ``SchedulerStopped`` event.

        """
        if self._state in (RunState.stopped, RunState.stopping):
            return

        event = anyio.Event()
        with self.event_broker.subscribe(
            lambda ev: event.set(), {SchedulerStopped}, one_shot=True
        ):
            await event.wait()

    async def _run(self, *, task_status: TaskStatus = TASK_STATUS_IGNORED) -> None:
        if self._state is not RunState.stopped:
            raise RuntimeError(
                f'Cannot start the scheduler when it is in the "{self._state}" '
                f"state"
            )

        self._state = RunState.starting
        async with AsyncExitStack() as exit_stack:
            self._wakeup_event = anyio.Event()

            # Initialize the event broker
            await self.event_broker.start()
            exit_stack.push_async_exit(
                lambda *exc_info: self.event_broker.stop(force=exc_info[0] is not None)
            )

            # Initialize the data store
            await self.data_store.start(self.event_broker)
            exit_stack.push_async_exit(
                lambda *exc_info: self.data_store.stop(force=exc_info[0] is not None)
            )

            # Wake up the scheduler if the data store emits a significant schedule event
            exit_stack.enter_context(
                self.event_broker.subscribe(
                    self._schedule_added_or_modified, {ScheduleAdded, ScheduleUpdated}
                )
            )

            # Start the built-in worker, if configured to do so
            if self.start_worker:
                token = current_scheduler.set(self)
                exit_stack.callback(current_scheduler.reset, token)
                worker = AsyncWorker(
                    self.data_store, self.event_broker, is_internal=True
                )
                await exit_stack.enter_async_context(worker)

            # Signal that the scheduler has started
            self._state = RunState.started
            task_status.started()
            await self.event_broker.publish_local(SchedulerStarted())

            exception: BaseException | None = None
            try:
                while self._state is RunState.started:
                    schedules = await self.data_store.acquire_schedules(
                        self.identity, 100
                    )
                    now = datetime.now(timezone.utc)
                    for schedule in schedules:
                        # Calculate a next fire time for the schedule, if possible
                        fire_times = [schedule.next_fire_time]
                        calculate_next = schedule.trigger.next
                        while True:
                            try:
                                fire_time = calculate_next()
                            except Exception:
                                self.logger.exception(
                                    "Error computing next fire time for schedule %r of "
                                    "task %r – removing schedule",
                                    schedule.id,
                                    schedule.task_id,
                                )
                                break

                            # Stop if the calculated fire time is in the future
                            if fire_time is None or fire_time > now:
                                schedule.next_fire_time = fire_time
                                break

                            # Only keep all the fire times if coalesce policy = "all"
                            if schedule.coalesce is CoalescePolicy.all:
                                fire_times.append(fire_time)
                            elif schedule.coalesce is CoalescePolicy.latest:
                                fire_times[0] = fire_time

                        # Add one or more jobs to the job queue
                        max_jitter = (
                            schedule.max_jitter.total_seconds()
                            if schedule.max_jitter
                            else 0
                        )
                        for i, fire_time in enumerate(fire_times):
                            # Calculate a jitter if max_jitter > 0
                            jitter = _zero_timedelta
                            if max_jitter:
                                if i + 1 < len(fire_times):
                                    next_fire_time = fire_times[i + 1]
                                else:
                                    next_fire_time = schedule.next_fire_time

                                if next_fire_time is not None:
                                    # Jitter must never be so high that it would cause a
                                    # fire time to equal or exceed the next fire time
                                    jitter_s = min(
                                        [
                                            max_jitter,
                                            (
                                                next_fire_time
                                                - fire_time
                                                - _microsecond_delta
                                            ).total_seconds(),
                                        ]
                                    )
                                    jitter = timedelta(
                                        seconds=random.uniform(0, jitter_s)
                                    )
                                    fire_time += jitter

                            schedule.last_fire_time = fire_time
                            job = Job(
                                task_id=schedule.task_id,
                                args=schedule.args,
                                kwargs=schedule.kwargs,
                                schedule_id=schedule.id,
                                scheduled_fire_time=fire_time,
                                jitter=jitter,
                                start_deadline=schedule.next_deadline,
                                tags=schedule.tags,
                            )
                            await self.data_store.add_job(job)

                    # Update the schedules (and release the scheduler's claim on them)
                    await self.data_store.release_schedules(self.identity, schedules)

                    # If we received fewer schedules than the maximum amount, sleep
                    # until the next schedule is due or the scheduler is explicitly
                    # woken up
                    wait_time = None
                    if len(schedules) < 100:
                        self._wakeup_deadline = (
                            await self.data_store.get_next_schedule_run_time()
                        )
                        if self._wakeup_deadline:
                            wait_time = (
                                self._wakeup_deadline - datetime.now(timezone.utc)
                            ).total_seconds()
                            self.logger.debug(
                                "Sleeping %.3f seconds until the next fire time (%s)",
                                wait_time,
                                self._wakeup_deadline,
                            )
                        else:
                            self.logger.debug("Waiting for any due schedules to appear")

                        with move_on_after(wait_time):
                            await self._wakeup_event.wait()
                            self._wakeup_event = anyio.Event()
                    else:
                        self.logger.debug(
                            "Processing more schedules on the next iteration"
                        )
            except BaseException as exc:
                exception = exc
                raise
            finally:
                self._state = RunState.stopped
                if isinstance(exception, Exception):
                    self.logger.exception("Scheduler crashed")
                elif exception:
                    self.logger.info(
                        f"Scheduler stopped due to {exception.__class__.__name__}"
                    )
                else:
                    self.logger.info("Scheduler stopped")

                with move_on_after(3, shield=True):
                    await self.event_broker.publish_local(
                        SchedulerStopped(exception=exception)
                    )



"""
Example demonstrating use of the synchronous scheduler.

To run: python sync_memory.py

It should print a line on the console on a one-second interval.
"""

from __future__ import annotations

from datetime import datetime

from apscheduler.schedulers.sync import Scheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


with Scheduler() as scheduler:
    scheduler.add_schedule(tick, IntervalTrigger(seconds=1))
    scheduler.wait_until_stopped()



"""
Example demonstrating use with ASGI (raw ASGI application, no framework).

Requires the "postgresql" service to be running.
To install prerequisites: pip install sqlalchemy asyncpg uvicorn
To run: uvicorn asgi_noframework:app

It should print a line on the console on a one-second interval while running a
basic web app at http://localhost:8000.
"""

from __future__ import annotations

from datetime import datetime

from sqlalchemy.ext.asyncio import create_async_engine

from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore
from apscheduler.eventbrokers.asyncpg import AsyncpgEventBroker
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


async def original_app(scope, receive, send):
    """Trivial example of an ASGI application."""
    if scope["type"] == "http":
        await receive()
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [
                    [b"content-type", b"text/plain"],
                ],
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": b"Hello, world!",
                "more_body": False,
            }
        )
    elif scope["type"] == "lifespan":
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                await send({"type": "lifespan.startup.complete"})
            elif message["type"] == "lifespan.shutdown":
                await send({"type": "lifespan.shutdown.complete"})
                return


async def scheduler_middleware(scope, receive, send):
    if scope["type"] == "lifespan":
        engine = create_async_engine(
            "postgresql+asyncpg://postgres:secret@localhost/testdb"
        )
        data_store = AsyncSQLAlchemyDataStore(engine)
        event_broker = AsyncpgEventBroker.from_async_sqla_engine(engine)
        async with AsyncScheduler(data_store, event_broker) as scheduler:
            await scheduler.add_schedule(tick, IntervalTrigger(seconds=1), id="tick")
            await original_app(scope, receive, send)
    else:
        await original_app(scope, receive, send)


# This is just for consistency with the other ASGI examples
app = scheduler_middleware



from __future__ import annotations

__all__ = [
    "CoalescePolicy",
    "ConflictPolicy",
    "ConflictingIdError",
    "DataStoreEvent",
    "DeserializationError",
    "Event",
    "Job",
    "JobAcquired",
    "JobAdded",
    "JobCancelled",
    "JobDeadlineMissed",
    "JobDeserializationFailed",
    "JobInfo",
    "JobLookupError",
    "JobOutcome",
    "JobReleased",
    "JobRemoved",
    "JobResult",
    "JobResultNotReady",
    "MaxIterationsReached",
    "RetrySettings",
    "RunState",
    "Schedule",
    "ScheduleLookupError",
    "SerializationError",
    "ScheduleAdded",
    "ScheduleUpdated",
    "ScheduleRemoved",
    "ScheduleDeserializationFailed",
    "SchedulerEvent",
    "SchedulerStarted",
    "SchedulerStopped",
    "Task",
    "TaskAdded",
    "TaskLookupError",
    "TaskUpdated",
    "TaskRemoved",
    "WorkerEvent",
    "WorkerStarted",
    "WorkerStopped",
    "current_scheduler",
    "current_worker",
    "current_job",
]

from typing import Any

from ._context import current_job, current_scheduler, current_worker
from ._enums import CoalescePolicy, ConflictPolicy, JobOutcome, RunState
from ._events import (
    DataStoreEvent,
    Event,
    JobAcquired,
    JobAdded,
    JobDeserializationFailed,
    JobReleased,
    JobRemoved,
    ScheduleAdded,
    ScheduleDeserializationFailed,
    ScheduleRemoved,
    SchedulerEvent,
    SchedulerStarted,
    SchedulerStopped,
    ScheduleUpdated,
    TaskAdded,
    TaskRemoved,
    TaskUpdated,
    WorkerEvent,
    WorkerStarted,
    WorkerStopped,
)
from ._exceptions import (
    ConflictingIdError,
    DeserializationError,
    JobCancelled,
    JobDeadlineMissed,
    JobLookupError,
    JobResultNotReady,
    MaxIterationsReached,
    ScheduleLookupError,
    SerializationError,
    TaskLookupError,
)
from ._structures import Job, JobInfo, JobResult, RetrySettings, Schedule, Task

# Re-export imports, so they look like they live directly in this package
value: Any
for value in list(locals().values()):
    if getattr(value, "__module__", "").startswith("apscheduler."):
        value.__module__ = __name__



from __future__ import annotations

from datetime import datetime, timedelta

import pytest

from apscheduler import MaxIterationsReached
from apscheduler.triggers.combining import AndTrigger, OrTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger


class TestAndTrigger:
    @pytest.mark.parametrize("threshold", [1, 0])
    def test_two_datetriggers(self, timezone, serializer, threshold):
        date1 = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        date2 = datetime(2020, 5, 16, 14, 17, 31, 254212, tzinfo=timezone)
        trigger = AndTrigger(
            [DateTrigger(date1), DateTrigger(date2)], threshold=threshold
        )
        if serializer:
            trigger = serializer.deserialize(serializer.serialize(trigger))

        if threshold:
            # date2 was within the threshold so it will not be used
            assert trigger.next() == date1

        assert trigger.next() is None

    def test_max_iterations(self, timezone, serializer):
        start_time = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        trigger = AndTrigger(
            [
                IntervalTrigger(seconds=4, start_time=start_time),
                IntervalTrigger(
                    seconds=4, start_time=start_time + timedelta(seconds=2)
                ),
            ]
        )
        if serializer:
            trigger = serializer.deserialize(serializer.serialize(trigger))

        pytest.raises(MaxIterationsReached, trigger.next)

    def test_repr(self, timezone, serializer):
        start_time = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        trigger = AndTrigger(
            [
                IntervalTrigger(seconds=4, start_time=start_time),
                IntervalTrigger(
                    seconds=4, start_time=start_time + timedelta(seconds=2)
                ),
            ]
        )
        if serializer:
            trigger = serializer.deserialize(serializer.serialize(trigger))

        assert repr(trigger) == (
            "AndTrigger([IntervalTrigger(seconds=4, "
            "start_time='2020-05-16 14:17:30.254212+02:00'), IntervalTrigger("
            "seconds=4, start_time='2020-05-16 14:17:32.254212+02:00')], "
            "threshold=1.0, max_iterations=10000)"
        )


class TestOrTrigger:
    def test_two_datetriggers(self, timezone, serializer):
        date1 = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        date2 = datetime(2020, 5, 18, 15, 1, 53, 940564, tzinfo=timezone)
        trigger = OrTrigger([DateTrigger(date1), DateTrigger(date2)])
        if serializer:
            trigger = serializer.deserialize(serializer.serialize(trigger))

        assert trigger.next() == date1
        assert trigger.next() == date2
        assert trigger.next() is None

    def test_two_interval_triggers(self, timezone, serializer):
        start_time = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        end_time1 = start_time + timedelta(seconds=16)
        end_time2 = start_time + timedelta(seconds=18)
        trigger = OrTrigger(
            [
                IntervalTrigger(seconds=4, start_time=start_time, end_time=end_time1),
                IntervalTrigger(seconds=6, start_time=start_time, end_time=end_time2),
            ]
        )
        if serializer:
            trigger = serializer.deserialize(serializer.serialize(trigger))

        assert trigger.next() == start_time
        assert trigger.next() == start_time + timedelta(seconds=4)
        assert trigger.next() == start_time + timedelta(seconds=6)
        assert trigger.next() == start_time + timedelta(seconds=8)
        assert trigger.next() == start_time + timedelta(seconds=12)
        assert trigger.next() == start_time + timedelta(seconds=16)
        # The end time of the 4 second interval has been reached
        assert trigger.next() == start_time + timedelta(seconds=18)
        # The end time of the 6 second interval has been reached
        assert trigger.next() is None

    def test_repr(self, timezone):
        date1 = datetime(2020, 5, 16, 14, 17, 30, 254212, tzinfo=timezone)
        date2 = datetime(2020, 5, 18, 15, 1, 53, 940564, tzinfo=timezone)
        trigger = OrTrigger([DateTrigger(date1), DateTrigger(date2)])
        assert repr(trigger) == (
            "OrTrigger([DateTrigger('2020-05-16 14:17:30.254212+02:00'), "
            "DateTrigger('2020-05-18 15:01:53.940564+02:00')])"
        )



"""
Example demonstrating use with the FastAPI web framework.

Requires the "postgresql" service to be running.
To install prerequisites: pip install sqlalchemy asycnpg fastapi uvicorn
To run: uvicorn asgi_fastapi:app

It should print a line on the console on a one-second interval while running a
basic web app at http://localhost:8000.
"""

from __future__ import annotations

from datetime import datetime

from fastapi import FastAPI
from fastapi.middleware import Middleware
from fastapi.requests import Request
from fastapi.responses import PlainTextResponse, Response
from sqlalchemy.ext.asyncio import create_async_engine
from starlette.types import ASGIApp, Receive, Scope, Send

from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore
from apscheduler.eventbrokers.asyncpg import AsyncpgEventBroker
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.triggers.interval import IntervalTrigger


def tick():
    print("Hello, the time is", datetime.now())


class SchedulerMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        scheduler: AsyncScheduler,
    ) -> None:
        self.app = app
        self.scheduler = scheduler

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":
            async with self.scheduler:
                await self.scheduler.add_schedule(
                    tick, IntervalTrigger(seconds=1), id="tick"
                )
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)


async def root(request: Request) -> Response:
    return PlainTextResponse("Hello, world!")


engine = create_async_engine("postgresql+asyncpg://postgres:secret@localhost/testdb")
data_store = AsyncSQLAlchemyDataStore(engine)
event_broker = AsyncpgEventBroker.from_async_sqla_engine(engine)
scheduler = AsyncScheduler()
middleware = [Middleware(SchedulerMiddleware, scheduler=scheduler)]
app = FastAPI(middleware=middleware)
app.add_api_route("/", root)



from __future__ import annotations

import sys
import threading
import time
from datetime import datetime, timedelta, timezone
from uuid import UUID

import anyio
import pytest
from anyio import fail_after
from pytest_mock import MockerFixture

from apscheduler import (
    Event,
    Job,
    JobAdded,
    JobLookupError,
    JobOutcome,
    JobReleased,
    Schedule,
    ScheduleAdded,
    ScheduleLookupError,
    ScheduleRemoved,
    SchedulerStopped,
    Task,
    TaskAdded,
    current_job,
    current_scheduler,
    current_worker,
)
from apscheduler.schedulers.async_ import AsyncScheduler
from apscheduler.schedulers.sync import Scheduler
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.workers.async_ import AsyncWorker

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo
else:
    from backports.zoneinfo import ZoneInfo

pytestmark = pytest.mark.anyio


async def dummy_async_job(delay: float = 0, fail: bool = False) -> str:
    await anyio.sleep(delay)
    if fail:
        raise RuntimeError("failing as requested")
    else:
        return "returnvalue"


def dummy_sync_job(delay: float = 0, fail: bool = False) -> str:
    time.sleep(delay)
    if fail:
        raise RuntimeError("failing as requested")
    else:
        return "returnvalue"


class TestAsyncScheduler:
    async def test_schedule_job(self) -> None:
        def listener(received_event: Event) -> None:
            received_events.append(received_event)
            if isinstance(received_event, ScheduleRemoved):
                event.set()

        received_events: list[Event] = []
        event = anyio.Event()
        trigger = DateTrigger(datetime.now(timezone.utc))
        async with AsyncScheduler(start_worker=False) as scheduler:
            scheduler.event_broker.subscribe(listener)
            await scheduler.add_schedule(dummy_async_job, trigger, id="foo")
            with fail_after(3):
                await event.wait()

        # Then the task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_schedulers:dummy_async_job"

        # Then a schedule was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, ScheduleAdded)
        assert received_event.schedule_id == "foo"
        # assert received_event.task_id == 'task_id'

        # Then that schedule was processed and a job was added for it
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.schedule_id == "foo"
        assert received_event.task_id == "test_schedulers:dummy_async_job"

        # Then the schedule was removed since the trigger had been exhausted
        received_event = received_events.pop(0)
        assert isinstance(received_event, ScheduleRemoved)
        assert received_event.schedule_id == "foo"

        # Finally, the scheduler was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, SchedulerStopped)

        # There should be no more events on the list
        assert not received_events

    async def test_add_get_schedule(self) -> None:
        async with AsyncScheduler(start_worker=False) as scheduler:
            with pytest.raises(ScheduleLookupError):
                await scheduler.get_schedule("dummyid")

            trigger = DateTrigger(datetime.now(timezone.utc) + timedelta(1))
            await scheduler.add_schedule(dummy_async_job, trigger, id="dummyid")
            schedule = await scheduler.get_schedule("dummyid")
            assert isinstance(schedule, Schedule)

    async def test_add_get_schedules(self) -> None:
        async with AsyncScheduler(start_worker=False) as scheduler:
            assert await scheduler.get_schedules() == []

            schedule1_id = await scheduler.add_schedule(
                dummy_async_job, DateTrigger(datetime.now() + timedelta(1))
            )
            await scheduler.add_schedule(
                dummy_async_job,
                DateTrigger(datetime.now() + timedelta(1)),
                id="dummyid",
            )

            schedules = await scheduler.get_schedules()
            assert all(isinstance(schedule, Schedule) for schedule in schedules)
            assert any(schedule.id == schedule1_id for schedule in schedules)
            assert any(schedule.id == "dummyid" for schedule in schedules)

    @pytest.mark.parametrize(
        "max_jitter, expected_upper_bound",
        [pytest.param(2, 2, id="within"), pytest.param(4, 2.999999, id="exceed")],
    )
    async def test_jitter(
        self,
        mocker: MockerFixture,
        timezone: ZoneInfo,
        max_jitter: float,
        expected_upper_bound: float,
    ) -> None:
        job_id: UUID | None = None

        def job_added_listener(event: Event) -> None:
            nonlocal job_id
            assert isinstance(event, JobAdded)
            job_id = event.job_id
            job_added_event.set()

        jitter = 1.569374
        orig_start_time = datetime.now(timezone) - timedelta(seconds=1)
        fake_uniform = mocker.patch("random.uniform")
        fake_uniform.configure_mock(side_effect=lambda a, b: jitter)
        async with AsyncScheduler(start_worker=False) as scheduler:
            trigger = IntervalTrigger(seconds=3, start_time=orig_start_time)
            job_added_event = anyio.Event()
            scheduler.event_broker.subscribe(job_added_listener, {JobAdded})
            schedule_id = await scheduler.add_schedule(
                dummy_async_job, trigger, max_jitter=max_jitter
            )
            schedule = await scheduler.get_schedule(schedule_id)
            assert schedule.max_jitter == timedelta(seconds=max_jitter)

            # Wait for the job to be added
            with fail_after(3):
                await job_added_event.wait()

            fake_uniform.assert_called_once_with(0, expected_upper_bound)

            # Check that the job was created with the proper amount of jitter in its
            # scheduled time
            jobs = await scheduler.data_store.get_jobs({job_id})
            assert jobs[0].jitter == timedelta(seconds=jitter)
            assert jobs[0].scheduled_fire_time == orig_start_time + timedelta(
                seconds=jitter
            )
            assert jobs[0].original_scheduled_time == orig_start_time

    async def test_get_job_result_success(self) -> None:
        async with AsyncScheduler() as scheduler:
            job_id = await scheduler.add_job(
                dummy_async_job, kwargs={"delay": 0.2}, result_expiration_time=5
            )
            result = await scheduler.get_job_result(job_id)
            assert result.job_id == job_id
            assert result.outcome is JobOutcome.success
            assert result.return_value == "returnvalue"

    async def test_get_job_result_success_empty(self) -> None:
        event = anyio.Event()
        async with AsyncScheduler() as scheduler:
            scheduler.event_broker.subscribe(
                lambda evt: event.set(), {JobReleased}, one_shot=True
            )
            job_id = await scheduler.add_job(dummy_async_job)
            with fail_after(3):
                await event.wait()

            with pytest.raises(JobLookupError):
                await scheduler.get_job_result(job_id, wait=False)

    async def test_get_job_result_error(self) -> None:
        async with AsyncScheduler() as scheduler:
            job_id = await scheduler.add_job(
                dummy_async_job,
                kwargs={"delay": 0.2, "fail": True},
                result_expiration_time=5,
            )
            result = await scheduler.get_job_result(job_id)
            assert result.job_id == job_id
            assert result.outcome is JobOutcome.error
            assert isinstance(result.exception, RuntimeError)
            assert str(result.exception) == "failing as requested"

    async def test_get_job_result_error_empty(self) -> None:
        event = anyio.Event()
        async with AsyncScheduler() as scheduler:
            scheduler.event_broker.subscribe(lambda evt: event.set(), one_shot=True)
            job_id = await scheduler.add_job(dummy_sync_job, kwargs={"fail": True})
            with fail_after(3):
                await event.wait()

            with pytest.raises(JobLookupError):
                await scheduler.get_job_result(job_id, wait=False)

    async def test_get_job_result_nowait_not_yet_ready(self) -> None:
        async with AsyncScheduler() as scheduler:
            job_id = await scheduler.add_job(dummy_async_job, kwargs={"delay": 0.2})
            with pytest.raises(JobLookupError):
                await scheduler.get_job_result(job_id, wait=False)

    async def test_run_job_success(self) -> None:
        async with AsyncScheduler() as scheduler:
            return_value = await scheduler.run_job(dummy_async_job)
            assert return_value == "returnvalue"

    async def test_run_job_failure(self) -> None:
        async with AsyncScheduler() as scheduler:
            with pytest.raises(RuntimeError, match="failing as requested"):
                await scheduler.run_job(dummy_async_job, kwargs={"fail": True})

    async def test_contextvars(self) -> None:
        def check_contextvars() -> None:
            assert current_scheduler.get() is scheduler
            assert isinstance(current_worker.get(), AsyncWorker)
            info = current_job.get()
            assert info.task_id == "task_id"
            assert info.schedule_id == "foo"
            assert info.scheduled_fire_time == scheduled_fire_time
            assert info.jitter == timedelta(seconds=2.16)
            assert info.start_deadline == start_deadline
            assert info.tags == {"foo", "bar"}

        scheduled_fire_time = datetime.now(timezone.utc)
        start_deadline = datetime.now(timezone.utc) + timedelta(seconds=10)
        async with AsyncScheduler() as scheduler:
            await scheduler.data_store.add_task(
                Task(id="task_id", func=check_contextvars)
            )
            job = Job(
                task_id="task_id",
                schedule_id="foo",
                scheduled_fire_time=scheduled_fire_time,
                jitter=timedelta(seconds=2.16),
                start_deadline=start_deadline,
                tags={"foo", "bar"},
                result_expiration_time=timedelta(seconds=10),
            )
            await scheduler.data_store.add_job(job)
            result = await scheduler.get_job_result(job.id)
            if result.outcome is JobOutcome.error:
                raise result.exception
            else:
                assert result.outcome is JobOutcome.success

    async def test_wait_until_stopped(self) -> None:
        async with AsyncScheduler() as scheduler:
            trigger = DateTrigger(
                datetime.now(timezone.utc) + timedelta(milliseconds=100)
            )
            await scheduler.add_schedule(scheduler.stop, trigger)
            await scheduler.wait_until_stopped()

        # This should be a no-op
        await scheduler.wait_until_stopped()


class TestSyncScheduler:
    def test_schedule_job(self):
        def listener(received_event: Event) -> None:
            received_events.append(received_event)
            if isinstance(received_event, ScheduleRemoved):
                event.set()

        received_events: list[Event] = []
        event = threading.Event()
        trigger = DateTrigger(datetime.now(timezone.utc))
        with Scheduler(start_worker=False) as scheduler:
            scheduler.event_broker.subscribe(listener)
            scheduler.add_schedule(dummy_sync_job, trigger, id="foo")
            event.wait(3)

        # First, a task was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_schedulers:dummy_sync_job"

        # Then a schedule was added
        received_event = received_events.pop(0)
        assert isinstance(received_event, ScheduleAdded)
        assert received_event.schedule_id == "foo"

        # Then that schedule was processed and a job was added for it
        received_event = received_events.pop(0)
        assert isinstance(received_event, JobAdded)
        assert received_event.schedule_id == "foo"
        assert received_event.task_id == "test_schedulers:dummy_sync_job"

        # Then the schedule was removed since the trigger had been exhausted
        received_event = received_events.pop(0)
        assert isinstance(received_event, ScheduleRemoved)
        assert received_event.schedule_id == "foo"

        # Finally, the scheduler was stopped
        received_event = received_events.pop(0)
        assert isinstance(received_event, SchedulerStopped)

    def test_add_get_schedule(self) -> None:
        with Scheduler(start_worker=False) as scheduler:
            with pytest.raises(ScheduleLookupError):
                scheduler.get_schedule("dummyid")

            trigger = DateTrigger(datetime.now(timezone.utc))
            scheduler.add_schedule(dummy_async_job, trigger, id="dummyid")
            schedule = scheduler.get_schedule("dummyid")
            assert isinstance(schedule, Schedule)

    def test_add_get_schedules(self) -> None:
        with Scheduler(start_worker=False) as scheduler:
            assert scheduler.get_schedules() == []

            schedule1_id = scheduler.add_schedule(
                dummy_sync_job, DateTrigger(datetime.now() + timedelta(1))
            )
            scheduler.add_schedule(
                dummy_sync_job, DateTrigger(datetime.now() + timedelta(1)), id="dummyid"
            )

            schedules = scheduler.get_schedules()
            assert all(isinstance(schedule, Schedule) for schedule in schedules)
            assert any(schedule.id == schedule1_id for schedule in schedules)
            assert any(schedule.id == "dummyid" for schedule in schedules)

    @pytest.mark.parametrize(
        "max_jitter, expected_upper_bound",
        [pytest.param(2, 2, id="within"), pytest.param(4, 2.999999, id="exceed")],
    )
    def test_jitter(
        self,
        mocker: MockerFixture,
        timezone: ZoneInfo,
        max_jitter: float,
        expected_upper_bound: float,
    ) -> None:
        job_id: UUID | None = None

        def job_added_listener(event: Event) -> None:
            nonlocal job_id
            assert isinstance(event, JobAdded)
            job_id = event.job_id
            job_added_event.set()

        jitter = 1.569374
        orig_start_time = datetime.now(timezone) - timedelta(seconds=1)
        fake_uniform = mocker.patch("random.uniform")
        fake_uniform.configure_mock(side_effect=lambda a, b: jitter)
        with Scheduler(start_worker=False) as scheduler:
            trigger = IntervalTrigger(seconds=3, start_time=orig_start_time)
            job_added_event = threading.Event()
            scheduler.event_broker.subscribe(job_added_listener, {JobAdded})
            schedule_id = scheduler.add_schedule(
                dummy_async_job, trigger, max_jitter=max_jitter
            )
            schedule = scheduler.get_schedule(schedule_id)
            assert schedule.max_jitter == timedelta(seconds=max_jitter)

            # Wait for the job to be added
            job_added_event.wait(3)

            fake_uniform.assert_called_once_with(0, expected_upper_bound)

            # Check that the job was created with the proper amount of jitter in its
            # scheduled time
            jobs = scheduler.data_store.get_jobs({job_id})
            assert jobs[0].jitter == timedelta(seconds=jitter)
            assert jobs[0].scheduled_fire_time == orig_start_time + timedelta(
                seconds=jitter
            )
            assert jobs[0].original_scheduled_time == orig_start_time

    def test_get_job_result_success(self) -> None:
        with Scheduler() as scheduler:
            job_id = scheduler.add_job(dummy_sync_job, result_expiration_time=5)
            result = scheduler.get_job_result(job_id)
            assert result.outcome is JobOutcome.success
            assert result.return_value == "returnvalue"

    def test_get_job_result_success_empty(self) -> None:
        event = threading.Event()
        with Scheduler() as scheduler:
            with scheduler.event_broker.subscribe(
                lambda evt: event.set(), {JobReleased}, one_shot=True
            ):
                job_id = scheduler.add_job(dummy_sync_job)
                event.wait(3)

            with pytest.raises(JobLookupError):
                scheduler.get_job_result(job_id, wait=False)

    def test_get_job_result_error(self) -> None:
        with Scheduler() as scheduler:
            job_id = scheduler.add_job(
                dummy_sync_job, kwargs={"fail": True}, result_expiration_time=5
            )
            result = scheduler.get_job_result(job_id)
            assert result.job_id == job_id
            assert result.outcome is JobOutcome.error
            assert isinstance(result.exception, RuntimeError)
            assert str(result.exception) == "failing as requested"

    def test_get_job_result_error_empty(self) -> None:
        event = threading.Event()
        with Scheduler() as scheduler, scheduler.event_broker.subscribe(
            lambda evt: event.set(), one_shot=True
        ):
            job_id = scheduler.add_job(dummy_sync_job, kwargs={"fail": True})
            event.wait(3)
            with pytest.raises(JobLookupError):
                scheduler.get_job_result(job_id, wait=False)

    def test_get_job_result_nowait_not_yet_ready(self) -> None:
        with Scheduler() as scheduler:
            job_id = scheduler.add_job(dummy_sync_job, kwargs={"delay": 0.2})
            with pytest.raises(JobLookupError):
                scheduler.get_job_result(job_id, wait=False)

    def test_run_job_success(self) -> None:
        with Scheduler() as scheduler:
            return_value = scheduler.run_job(dummy_sync_job)
            assert return_value == "returnvalue"

    def test_run_job_failure(self) -> None:
        with Scheduler() as scheduler:
            with pytest.raises(RuntimeError, match="failing as requested"):
                scheduler.run_job(dummy_sync_job, kwargs={"fail": True})

    def test_contextvars(self) -> None:
        def check_contextvars() -> None:
            assert current_scheduler.get() is scheduler
            assert current_worker.get() is not None
            info = current_job.get()
            assert info.task_id == "task_id"
            assert info.schedule_id == "foo"
            assert info.scheduled_fire_time == scheduled_fire_time
            assert info.jitter == timedelta(seconds=2.16)
            assert info.start_deadline == start_deadline
            assert info.tags == {"foo", "bar"}

        scheduled_fire_time = datetime.now(timezone.utc)
        start_deadline = datetime.now(timezone.utc) + timedelta(seconds=10)
        with Scheduler() as scheduler:
            scheduler.data_store.add_task(Task(id="task_id", func=check_contextvars))
            job = Job(
                task_id="task_id",
                schedule_id="foo",
                scheduled_fire_time=scheduled_fire_time,
                jitter=timedelta(seconds=2.16),
                start_deadline=start_deadline,
                tags={"foo", "bar"},
                result_expiration_time=timedelta(seconds=10),
            )
            scheduler.data_store.add_job(job)
            result = scheduler.get_job_result(job.id)
            if result.outcome is JobOutcome.error:
                raise result.exception
            else:
                assert result.outcome is JobOutcome.success

    def test_wait_until_stopped(self) -> None:
        with Scheduler() as scheduler:
            trigger = DateTrigger(
                datetime.now(timezone.utc) + timedelta(milliseconds=100)
            )
            scheduler.add_schedule(scheduler.stop, trigger)
            scheduler.wait_until_stopped()

        # This should be a no-op
        scheduler.wait_until_stopped()



"""
Example demonstrating the separation of scheduler and worker.
This script runs the worker part. You need to be running both this and the scheduler
script simultaneously in order for the scheduled task to be run.

Requires the "postgresql" and "redis" services to be running.
To install prerequisites: pip install sqlalchemy psycopg2 redis
To run: python sync_worker.py

When run together with sync_scheduler.py, it should print a line on the
console on a one-second interval.
"""

from __future__ import annotations

from sqlalchemy.future import create_engine

from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore
from apscheduler.eventbrokers.redis import RedisEventBroker
from apscheduler.workers.sync import Worker

engine = create_engine("postgresql+psycopg2://postgres:secret@localhost/testdb")
data_store = SQLAlchemyDataStore(engine)
event_broker = RedisEventBroker.from_url("redis://localhost")
worker = Worker(data_store, event_broker)
worker.run_until_stopped()



from __future__ import annotations

from collections.abc import Callable
from datetime import datetime, timedelta
from enum import Enum
from typing import TypeVar
from uuid import UUID

from . import abc

TEnum = TypeVar("TEnum", bound=Enum)


def as_aware_datetime(value: datetime | str) -> datetime:
    """Convert the value from a string to a timezone aware datetime."""
    if isinstance(value, str):
        # fromisoformat() does not handle the "Z" suffix
        if value.upper().endswith("Z"):
            value = value[:-1] + "+00:00"

        value = datetime.fromisoformat(value)

    return value


def as_uuid(value: UUID | str) -> UUID:
    """Convert a string-formatted UUID to a UUID instance."""
    if isinstance(value, str):
        return UUID(value)

    return value


def as_timedelta(value: timedelta | float | None) -> timedelta | None:
    if isinstance(value, (float, int)):
        return timedelta(seconds=value)

    return value


def as_enum(enum_class: type[TEnum]) -> Callable[[TEnum | str], TEnum]:
    def converter(value: TEnum | str) -> TEnum:
        if isinstance(value, str):
            return enum_class.__members__[value]

        return value

    return converter


def as_async_eventbroker(
    value: abc.EventBroker | abc.AsyncEventBroker,
) -> abc.AsyncEventBroker:
    if isinstance(value, abc.EventBroker):
        from apscheduler.eventbrokers.async_adapter import AsyncEventBrokerAdapter

        return AsyncEventBrokerAdapter(value)

    return value


def as_async_datastore(value: abc.DataStore | abc.AsyncDataStore) -> abc.AsyncDataStore:
    if isinstance(value, abc.DataStore):
        from apscheduler.datastores.async_adapter import AsyncDataStoreAdapter

        return AsyncDataStoreAdapter(value)

    return value



from __future__ import annotations

import sys
from datetime import datetime
from typing import Iterable
from uuid import UUID

import attrs
from anyio import to_thread
from anyio.from_thread import BlockingPortal

from .._enums import ConflictPolicy
from .._structures import Job, JobResult, Schedule, Task
from ..abc import AsyncEventBroker, DataStore
from ..eventbrokers.async_adapter import AsyncEventBrokerAdapter, SyncEventBrokerAdapter
from .base import BaseAsyncDataStore


@attrs.define(eq=False)
class AsyncDataStoreAdapter(BaseAsyncDataStore):
    original: DataStore
    _portal: BlockingPortal = attrs.field(init=False)

    async def start(self, event_broker: AsyncEventBroker) -> None:
        await super().start(event_broker)

        self._portal = BlockingPortal()
        await self._portal.__aenter__()

        if isinstance(event_broker, AsyncEventBrokerAdapter):
            sync_event_broker = event_broker.original
        else:
            sync_event_broker = SyncEventBrokerAdapter(event_broker, self._portal)

        try:
            await to_thread.run_sync(lambda: self.original.start(sync_event_broker))
        except BaseException:
            await self._portal.__aexit__(*sys.exc_info())
            raise

    async def stop(self, *, force: bool = False) -> None:
        try:
            await to_thread.run_sync(lambda: self.original.stop(force=force))
        finally:
            await self._portal.__aexit__(None, None, None)
            await super().stop(force=force)

    async def add_task(self, task: Task) -> None:
        await to_thread.run_sync(self.original.add_task, task)

    async def remove_task(self, task_id: str) -> None:
        await to_thread.run_sync(self.original.remove_task, task_id)

    async def get_task(self, task_id: str) -> Task:
        return await to_thread.run_sync(self.original.get_task, task_id)

    async def get_tasks(self) -> list[Task]:
        return await to_thread.run_sync(self.original.get_tasks)

    async def get_schedules(self, ids: set[str] | None = None) -> list[Schedule]:
        return await to_thread.run_sync(self.original.get_schedules, ids)

    async def add_schedule(
        self, schedule: Schedule, conflict_policy: ConflictPolicy
    ) -> None:
        await to_thread.run_sync(self.original.add_schedule, schedule, conflict_policy)

    async def remove_schedules(self, ids: Iterable[str]) -> None:
        await to_thread.run_sync(self.original.remove_schedules, ids)

    async def acquire_schedules(self, scheduler_id: str, limit: int) -> list[Schedule]:
        return await to_thread.run_sync(
            self.original.acquire_schedules, scheduler_id, limit
        )

    async def release_schedules(
        self, scheduler_id: str, schedules: list[Schedule]
    ) -> None:
        await to_thread.run_sync(
            self.original.release_schedules, scheduler_id, schedules
        )

    async def get_next_schedule_run_time(self) -> datetime | None:
        return await to_thread.run_sync(self.original.get_next_schedule_run_time)

    async def add_job(self, job: Job) -> None:
        await to_thread.run_sync(self.original.add_job, job)

    async def get_jobs(self, ids: Iterable[UUID] | None = None) -> list[Job]:
        return await to_thread.run_sync(self.original.get_jobs, ids)

    async def acquire_jobs(self, worker_id: str, limit: int | None = None) -> list[Job]:
        return await to_thread.run_sync(self.original.acquire_jobs, worker_id, limit)

    async def release_job(
        self, worker_id: str, task_id: str, result: JobResult
    ) -> None:
        await to_thread.run_sync(self.original.release_job, worker_id, task_id, result)

    async def get_job_result(self, job_id: UUID) -> JobResult | None:
        return await to_thread.run_sync(self.original.get_job_result, job_id)



from __future__ import annotations

from uuid import uuid4

import pytest

from apscheduler import Event, JobAdded
from apscheduler.abc import Serializer


@pytest.mark.parametrize(
    "event",
    [
        pytest.param(
            JobAdded(
                job_id=uuid4(),
                task_id="task",
                schedule_id="schedule",
                tags=["tag1", "tag2"],
            ),
            id="job_added",
        )
    ],
)
def test_serialize_event(event: Event, serializer: Serializer) -> None:
    payload = serializer.serialize(event.marshal(serializer))
    deserialized = type(event).unmarshal(serializer, serializer.deserialize(payload))
    assert deserialized == event



from __future__ import annotations

import operator
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from logging import Logger, getLogger
from typing import Any, Callable, ClassVar, Iterable
from uuid import UUID

import attrs
import pymongo
import tenacity
from attrs.validators import instance_of
from bson import CodecOptions, UuidRepresentation
from bson.codec_options import TypeEncoder, TypeRegistry
from pymongo import ASCENDING, DeleteOne, MongoClient, UpdateOne
from pymongo.collection import Collection
from pymongo.errors import ConnectionFailure, DuplicateKeyError

from .._enums import CoalescePolicy, ConflictPolicy, JobOutcome
from .._events import (
    DataStoreEvent,
    JobAcquired,
    JobAdded,
    ScheduleAdded,
    ScheduleRemoved,
    ScheduleUpdated,
    TaskAdded,
    TaskRemoved,
    TaskUpdated,
)
from .._exceptions import (
    ConflictingIdError,
    DeserializationError,
    SerializationError,
    TaskLookupError,
)
from .._structures import Job, JobResult, RetrySettings, Schedule, Task
from ..abc import EventBroker, Serializer
from ..serializers.pickle import PickleSerializer
from .base import BaseDataStore


class CustomEncoder(TypeEncoder):
    def __init__(self, python_type: type, encoder: Callable):
        self._python_type = python_type
        self._encoder = encoder

    @property
    def python_type(self) -> type:
        return self._python_type

    def transform_python(self, value: Any) -> Any:
        return self._encoder(value)


@attrs.define(eq=False)
class MongoDBDataStore(BaseDataStore):
    """
    Uses a MongoDB server to store data.

    When started, this data store creates the appropriate indexes on the given database
    if they're not already present.

    Operations are retried (in accordance to ``retry_settings``) when an operation
    raises :exc:`pymongo.errors.ConnectionFailure`.

    :param client: a PyMongo client
    :param serializer: the serializer used to (de)serialize tasks, schedules and jobs
        for storage
    :param database: name of the database to use
    :param lock_expiration_delay: maximum amount of time (in seconds) that a scheduler
        or worker can keep a lock on a schedule or task
    :param retry_settings: Tenacity settings for retrying operations in case of a
        database connecitivty problem
    :param start_from_scratch: erase all existing data during startup (useful for test
        suites)
    """

    client: MongoClient = attrs.field(validator=instance_of(MongoClient))
    serializer: Serializer = attrs.field(factory=PickleSerializer, kw_only=True)
    database: str = attrs.field(default="apscheduler", kw_only=True)
    lock_expiration_delay: float = attrs.field(default=30, kw_only=True)
    retry_settings: RetrySettings = attrs.field(default=RetrySettings(), kw_only=True)
    start_from_scratch: bool = attrs.field(default=False, kw_only=True)

    _task_attrs: ClassVar[list[str]] = [field.name for field in attrs.fields(Task)]
    _schedule_attrs: ClassVar[list[str]] = [
        field.name for field in attrs.fields(Schedule)
    ]
    _job_attrs: ClassVar[list[str]] = [field.name for field in attrs.fields(Job)]

    _logger: Logger = attrs.field(init=False, factory=lambda: getLogger(__name__))
    _local_tasks: dict[str, Task] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self) -> None:
        type_registry = TypeRegistry(
            [
                CustomEncoder(timedelta, timedelta.total_seconds),
                CustomEncoder(ConflictPolicy, operator.attrgetter("name")),
                CustomEncoder(CoalescePolicy, operator.attrgetter("name")),
                CustomEncoder(JobOutcome, operator.attrgetter("name")),
            ]
        )
        codec_options = CodecOptions(
            tz_aware=True,
            type_registry=type_registry,
            uuid_representation=UuidRepresentation.STANDARD,
        )
        database = self.client.get_database(self.database, codec_options=codec_options)
        self._tasks: Collection = database["tasks"]
        self._schedules: Collection = database["schedules"]
        self._jobs: Collection = database["jobs"]
        self._jobs_results: Collection = database["job_results"]

    @classmethod
    def from_url(cls, uri: str, **options) -> MongoDBDataStore:
        client = MongoClient(uri)
        return cls(client, **options)

    def _retry(self) -> tenacity.Retrying:
        return tenacity.Retrying(
            stop=self.retry_settings.stop,
            wait=self.retry_settings.wait,
            retry=tenacity.retry_if_exception_type(ConnectionFailure),
            after=self._after_attempt,
            reraise=True,
        )

    def _after_attempt(self, retry_state: tenacity.RetryCallState) -> None:
        self._logger.warning(
            "Temporary data store error (attempt %d): %s",
            retry_state.attempt_number,
            retry_state.outcome.exception(),
        )

    def start(self, event_broker: EventBroker) -> None:
        super().start(event_broker)

        server_info = self.client.server_info()
        if server_info["versionArray"] < [4, 0]:
            raise RuntimeError(
                f"MongoDB server must be at least v4.0; current version = "
                f"{server_info['version']}"
            )

        for attempt in self._retry():
            with attempt, self.client.start_session() as session:
                if self.start_from_scratch:
                    self._tasks.delete_many({}, session=session)
                    self._schedules.delete_many({}, session=session)
                    self._jobs.delete_many({}, session=session)
                    self._jobs_results.delete_many({}, session=session)

                self._schedules.create_index("next_fire_time", session=session)
                self._jobs.create_index("task_id", session=session)
                self._jobs.create_index("created_at", session=session)
                self._jobs.create_index("tags", session=session)
                self._jobs_results.create_index("finished_at", session=session)
                self._jobs_results.create_index("expires_at", session=session)

    def add_task(self, task: Task) -> None:
        for attempt in self._retry():
            with attempt:
                previous = self._tasks.find_one_and_update(
                    {"_id": task.id},
                    {
                        "$set": task.marshal(self.serializer),
                        "$setOnInsert": {"running_jobs": 0},
                    },
                    upsert=True,
                )

        self._local_tasks[task.id] = task
        if previous:
            self._events.publish(TaskUpdated(task_id=task.id))
        else:
            self._events.publish(TaskAdded(task_id=task.id))

    def remove_task(self, task_id: str) -> None:
        for attempt in self._retry():
            with attempt:
                if not self._tasks.find_one_and_delete({"_id": task_id}):
                    raise TaskLookupError(task_id)

        del self._local_tasks[task_id]
        self._events.publish(TaskRemoved(task_id=task_id))

    def get_task(self, task_id: str) -> Task:
        try:
            return self._local_tasks[task_id]
        except KeyError:
            for attempt in self._retry():
                with attempt:
                    document = self._tasks.find_one(
                        {"_id": task_id}, projection=self._task_attrs
                    )

            if not document:
                raise TaskLookupError(task_id)

            document["id"] = document.pop("id")
            task = self._local_tasks[task_id] = Task.unmarshal(
                self.serializer, document
            )
            return task

    def get_tasks(self) -> list[Task]:
        for attempt in self._retry():
            with attempt:
                tasks: list[Task] = []
                for document in self._tasks.find(
                    projection=self._task_attrs, sort=[("_id", pymongo.ASCENDING)]
                ):
                    document["id"] = document.pop("_id")
                    tasks.append(Task.unmarshal(self.serializer, document))

        return tasks

    def get_schedules(self, ids: set[str] | None = None) -> list[Schedule]:
        filters = {"_id": {"$in": list(ids)}} if ids is not None else {}
        for attempt in self._retry():
            with attempt:
                schedules: list[Schedule] = []
                cursor = self._schedules.find(filters).sort("_id")
                for document in cursor:
                    document["id"] = document.pop("_id")
                    try:
                        schedule = Schedule.unmarshal(self.serializer, document)
                    except DeserializationError:
                        self._logger.warning(
                            "Failed to deserialize schedule %r", document["_id"]
                        )
                        continue

                    schedules.append(schedule)

        return schedules

    def add_schedule(self, schedule: Schedule, conflict_policy: ConflictPolicy) -> None:
        event: DataStoreEvent
        document = schedule.marshal(self.serializer)
        document["_id"] = document.pop("id")
        try:
            for attempt in self._retry():
                with attempt:
                    self._schedules.insert_one(document)
        except DuplicateKeyError:
            if conflict_policy is ConflictPolicy.exception:
                raise ConflictingIdError(schedule.id) from None
            elif conflict_policy is ConflictPolicy.replace:
                for attempt in self._retry():
                    with attempt:
                        self._schedules.replace_one(
                            {"_id": schedule.id}, document, True
                        )

                event = ScheduleUpdated(
                    schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
                )
                self._events.publish(event)
        else:
            event = ScheduleAdded(
                schedule_id=schedule.id, next_fire_time=schedule.next_fire_time
            )
            self._events.publish(event)

    def remove_schedules(self, ids: Iterable[str]) -> None:
        filters = {"_id": {"$in": list(ids)}} if ids is not None else {}
        for attempt in self._retry():
            with attempt, self.client.start_session() as session:
                cursor = self._schedules.find(
                    filters, projection=["_id"], session=session
                )
                ids = [doc["_id"] for doc in cursor]
                if ids:
                    self._schedules.delete_many(filters, session=session)

        for schedule_id in ids:
            self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    def acquire_schedules(self, scheduler_id: str, limit: int) -> list[Schedule]:
        for attempt in self._retry():
            with attempt, self.client.start_session() as session:
                schedules: list[Schedule] = []
                cursor = (
                    self._schedules.find(
                        {
                            "next_fire_time": {"$ne": None},
                            "$or": [
                                {"acquired_until": {"$exists": False}},
                                {"acquired_until": {"$lt": datetime.now(timezone.utc)}},
                            ],
                        },
                        session=session,
                    )
                    .sort("next_fire_time")
                    .limit(limit)
                )
                for document in cursor:
                    document["id"] = document.pop("_id")
                    schedule = Schedule.unmarshal(self.serializer, document)
                    schedules.append(schedule)

                if schedules:
                    now = datetime.now(timezone.utc)
                    acquired_until = datetime.fromtimestamp(
                        now.timestamp() + self.lock_expiration_delay, now.tzinfo
                    )
                    filters = {"_id": {"$in": [schedule.id for schedule in schedules]}}
                    update = {
                        "$set": {
                            "acquired_by": scheduler_id,
                            "acquired_until": acquired_until,
                        }
                    }
                    self._schedules.update_many(filters, update, session=session)

        return schedules

    def release_schedules(self, scheduler_id: str, schedules: list[Schedule]) -> None:
        updated_schedules: list[tuple[str, datetime]] = []
        finished_schedule_ids: list[str] = []

        # Update schedules that have a next fire time
        requests = []
        for schedule in schedules:
            filters = {"_id": schedule.id, "acquired_by": scheduler_id}
            if schedule.next_fire_time is not None:
                try:
                    serialized_trigger = self.serializer.serialize(schedule.trigger)
                except SerializationError:
                    self._logger.exception(
                        "Error serializing schedule %r – " "removing from data store",
                        schedule.id,
                    )
                    requests.append(DeleteOne(filters))
                    finished_schedule_ids.append(schedule.id)
                    continue

                update = {
                    "$unset": {
                        "acquired_by": True,
                        "acquired_until": True,
                    },
                    "$set": {
                        "trigger": serialized_trigger,
                        "next_fire_time": schedule.next_fire_time,
                    },
                }
                requests.append(UpdateOne(filters, update))
                updated_schedules.append((schedule.id, schedule.next_fire_time))
            else:
                requests.append(DeleteOne(filters))
                finished_schedule_ids.append(schedule.id)

            if requests:
                for attempt in self._retry():
                    with attempt, self.client.start_session() as session:
                        self._schedules.bulk_write(
                            requests, ordered=False, session=session
                        )

        for schedule_id, next_fire_time in updated_schedules:
            event = ScheduleUpdated(
                schedule_id=schedule_id, next_fire_time=next_fire_time
            )
            self._events.publish(event)

        for schedule_id in finished_schedule_ids:
            self._events.publish(ScheduleRemoved(schedule_id=schedule_id))

    def get_next_schedule_run_time(self) -> datetime | None:
        for attempt in self._retry():
            with attempt:
                document = self._schedules.find_one(
                    {"next_run_time": {"$ne": None}},
                    projection=["next_run_time"],
                    sort=[("next_run_time", ASCENDING)],
                )

        if document:
            return document["next_run_time"]
        else:
            return None

    def add_job(self, job: Job) -> None:
        document = job.marshal(self.serializer)
        document["_id"] = document.pop("id")
        for attempt in self._retry():
            with attempt:
                self._jobs.insert_one(document)

        event = JobAdded(
            job_id=job.id,
            task_id=job.task_id,
            schedule_id=job.schedule_id,
            tags=job.tags,
        )
        self._events.publish(event)

    def get_jobs(self, ids: Iterable[UUID] | None = None) -> list[Job]:
        filters = {"_id": {"$in": list(ids)}} if ids is not None else {}
        for attempt in self._retry():
            with attempt:
                jobs: list[Job] = []
                cursor = self._jobs.find(filters).sort("_id")
                for document in cursor:
                    document["id"] = document.pop("_id")
                    try:
                        job = Job.unmarshal(self.serializer, document)
                    except DeserializationError:
                        self._logger.warning(
                            "Failed to deserialize job %r", document["id"]
                        )
                        continue

                    jobs.append(job)

        return jobs

    def acquire_jobs(self, worker_id: str, limit: int | None = None) -> list[Job]:
        for attempt in self._retry():
            with attempt, self.client.start_session() as session:
                cursor = self._jobs.find(
                    {
                        "$or": [
                            {"acquired_until": {"$exists": False}},
                            {"acquired_until": {"$lt": datetime.now(timezone.utc)}},
                        ]
                    },
                    sort=[("created_at", ASCENDING)],
                    limit=limit,
                    session=session,
                )
                documents = list(cursor)

                # Retrieve the limits
                task_ids: set[str] = {document["task_id"] for document in documents}
                task_limits = self._tasks.find(
                    {"_id": {"$in": list(task_ids)}, "max_running_jobs": {"$ne": None}},
                    projection=["max_running_jobs", "running_jobs"],
                    session=session,
                )
                job_slots_left = {
                    doc["_id"]: doc["max_running_jobs"] - doc["running_jobs"]
                    for doc in task_limits
                }

                # Filter out jobs that don't have free slots
                acquired_jobs: list[Job] = []
                increments: dict[str, int] = defaultdict(lambda: 0)
                for document in documents:
                    document["id"] = document.pop("_id")
                    job = Job.unmarshal(self.serializer, document)

                    # Don't acquire the job if there are no free slots left
                    slots_left = job_slots_left.get(job.task_id)
                    if slots_left == 0:
                        continue
                    elif slots_left is not None:
                        job_slots_left[job.task_id] -= 1

                    acquired_jobs.append(job)
                    increments[job.task_id] += 1

                if acquired_jobs:
                    now = datetime.now(timezone.utc)
                    acquired_until = datetime.fromtimestamp(
                        now.timestamp() + self.lock_expiration_delay, timezone.utc
                    )
                    filters = {"_id": {"$in": [job.id for job in acquired_jobs]}}
                    update = {
                        "$set": {
                            "acquired_by": worker_id,
                            "acquired_until": acquired_until,
                        }
                    }
                    self._jobs.update_many(filters, update, session=session)

                    # Increment the running job counters on each task
                    for task_id, increment in increments.items():
                        self._tasks.find_one_and_update(
                            {"_id": task_id},
                            {"$inc": {"running_jobs": increment}},
                            session=session,
                        )

        # Publish the appropriate events
        for job in acquired_jobs:
            self._events.publish(JobAcquired(job_id=job.id, worker_id=worker_id))

        return acquired_jobs

    def release_job(self, worker_id: str, task_id: str, result: JobResult) -> None:
        for attempt in self._retry():
            with attempt, self.client.start_session() as session:
                # Record the job result
                if result.expires_at > result.finished_at:
                    document = result.marshal(self.serializer)
                    document["_id"] = document.pop("job_id")
                    self._jobs_results.insert_one(document, session=session)

                # Decrement the running jobs counter
                self._tasks.find_one_and_update(
                    {"_id": task_id}, {"$inc": {"running_jobs": -1}}, session=session
                )

                # Delete the job
                self._jobs.delete_one({"_id": result.job_id}, session=session)

    def get_job_result(self, job_id: UUID) -> JobResult | None:
        for attempt in self._retry():
            with attempt:
                document = self._jobs_results.find_one_and_delete({"_id": job_id})

        if document:
            document["job_id"] = document.pop("_id")
            return JobResult.unmarshal(self.serializer, document)
        else:
            return None



from __future__ import annotations

import threading
from collections.abc import Generator
from contextlib import asynccontextmanager, contextmanager
from datetime import datetime, timedelta, timezone
from tempfile import TemporaryDirectory
from typing import Any, AsyncGenerator, cast

import anyio
import pytest
from _pytest.fixtures import SubRequest
from anyio import CancelScope
from freezegun.api import FrozenDateTimeFactory
from pytest_lazyfixture import lazy_fixture

from apscheduler import (
    CoalescePolicy,
    ConflictPolicy,
    Event,
    JobOutcome,
    ScheduleAdded,
    ScheduleRemoved,
    ScheduleUpdated,
    TaskAdded,
    TaskLookupError,
    TaskUpdated,
)
from apscheduler._structures import Job, JobResult, Schedule, Task
from apscheduler.abc import AsyncDataStore, AsyncEventBroker, DataStore, EventBroker
from apscheduler.datastores.async_adapter import AsyncDataStoreAdapter
from apscheduler.datastores.memory import MemoryDataStore
from apscheduler.eventbrokers.async_local import LocalAsyncEventBroker
from apscheduler.eventbrokers.local import LocalEventBroker
from apscheduler.triggers.date import DateTrigger


@pytest.fixture
def memory_store() -> DataStore:
    yield MemoryDataStore()


@pytest.fixture
def adapted_memory_store() -> AsyncDataStore:
    store = MemoryDataStore()
    return AsyncDataStoreAdapter(store)


@pytest.fixture
def mongodb_store() -> DataStore:
    from pymongo import MongoClient

    from apscheduler.datastores.mongodb import MongoDBDataStore

    with MongoClient(tz_aware=True, serverSelectionTimeoutMS=1000) as client:
        yield MongoDBDataStore(client, start_from_scratch=True)


@pytest.fixture
def sqlite_store() -> DataStore:
    from sqlalchemy.future import create_engine

    from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore

    with TemporaryDirectory("sqlite_") as tempdir:
        engine = create_engine(f"sqlite:///{tempdir}/test.db")
        try:
            yield SQLAlchemyDataStore(engine)
        finally:
            engine.dispose()


@pytest.fixture
def psycopg2_store() -> DataStore:
    from sqlalchemy.future import create_engine

    from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore

    engine = create_engine("postgresql+psycopg2://postgres:secret@localhost/testdb")
    try:
        yield SQLAlchemyDataStore(engine, start_from_scratch=True)
    finally:
        engine.dispose()


@pytest.fixture
def mysql_store() -> DataStore:
    from sqlalchemy.future import create_engine

    from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore

    engine = create_engine("mysql+pymysql://root:secret@localhost/testdb")
    try:
        yield SQLAlchemyDataStore(engine, start_from_scratch=True)
    finally:
        engine.dispose()


@pytest.fixture
async def asyncpg_store() -> AsyncDataStore:
    from sqlalchemy.ext.asyncio import create_async_engine

    from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore

    engine = create_async_engine(
        "postgresql+asyncpg://postgres:secret@localhost/testdb", future=True
    )
    try:
        yield AsyncSQLAlchemyDataStore(engine, start_from_scratch=True)
    finally:
        await engine.dispose()


@pytest.fixture
async def asyncmy_store() -> AsyncDataStore:
    from sqlalchemy.ext.asyncio import create_async_engine

    from apscheduler.datastores.async_sqlalchemy import AsyncSQLAlchemyDataStore

    engine = create_async_engine(
        "mysql+asyncmy://root:secret@localhost/testdb?charset=utf8mb4", future=True
    )
    try:
        yield AsyncSQLAlchemyDataStore(engine, start_from_scratch=True)
    finally:
        await engine.dispose()


@pytest.fixture
def schedules() -> list[Schedule]:
    trigger = DateTrigger(datetime(2020, 9, 13, tzinfo=timezone.utc))
    schedule1 = Schedule(id="s1", task_id="task1", trigger=trigger)
    schedule1.next_fire_time = trigger.next()

    trigger = DateTrigger(datetime(2020, 9, 14, tzinfo=timezone.utc))
    schedule2 = Schedule(id="s2", task_id="task2", trigger=trigger)
    schedule2.next_fire_time = trigger.next()

    trigger = DateTrigger(datetime(2020, 9, 15, tzinfo=timezone.utc))
    schedule3 = Schedule(id="s3", task_id="task1", trigger=trigger)
    return [schedule1, schedule2, schedule3]


class TestDataStores:
    @contextmanager
    def capture_events(
        self,
        datastore: DataStore,
        limit: int,
        event_types: set[type[Event]] | None = None,
    ) -> Generator[list[Event], None, None]:
        def listener(event: Event) -> None:
            events.append(event)
            if len(events) == limit:
                limit_event.set()
                subscription.unsubscribe()

        events: list[Event] = []
        limit_event = threading.Event()
        subscription = datastore.events.subscribe(listener, event_types)
        yield events
        if limit:
            limit_event.wait(2)

    @pytest.fixture
    def event_broker(self) -> Generator[EventBroker, Any, None]:
        broker = LocalEventBroker()
        broker.start()
        yield broker
        broker.stop()

    @pytest.fixture(
        params=[
            pytest.param(lazy_fixture("memory_store"), id="memory"),
            pytest.param(lazy_fixture("sqlite_store"), id="sqlite"),
            pytest.param(
                lazy_fixture("mongodb_store"),
                id="mongodb",
                marks=[pytest.mark.external_service],
            ),
            pytest.param(
                lazy_fixture("psycopg2_store"),
                id="psycopg2",
                marks=[pytest.mark.external_service],
            ),
            pytest.param(
                lazy_fixture("mysql_store"),
                id="mysql",
                marks=[pytest.mark.external_service],
            ),
        ]
    )
    def datastore(
        self, request: SubRequest, event_broker: EventBroker
    ) -> Generator[DataStore, Any, None]:
        datastore = cast(DataStore, request.param)
        datastore.start(event_broker)
        yield datastore
        datastore.stop()

    def test_add_replace_task(self, datastore: DataStore) -> None:
        import math

        event_types = {TaskAdded, TaskUpdated}
        with self.capture_events(datastore, 3, event_types) as events:
            datastore.add_task(Task(id="test_task", func=print))
            datastore.add_task(Task(id="test_task2", func=math.ceil))
            datastore.add_task(Task(id="test_task", func=repr))

            tasks = datastore.get_tasks()
            assert len(tasks) == 2
            assert tasks[0].id == "test_task"
            assert tasks[0].func is repr
            assert tasks[1].id == "test_task2"
            assert tasks[1].func is math.ceil

        received_event = events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_task"

        received_event = events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_task2"

        received_event = events.pop(0)
        assert isinstance(received_event, TaskUpdated)
        assert received_event.task_id == "test_task"

        assert not events

    def test_add_schedules(
        self, datastore: DataStore, schedules: list[Schedule]
    ) -> None:
        with self.capture_events(datastore, 3, {ScheduleAdded}) as events:
            for schedule in schedules:
                datastore.add_schedule(schedule, ConflictPolicy.exception)

            assert datastore.get_schedules() == schedules
            assert datastore.get_schedules({"s1", "s2", "s3"}) == schedules
            assert datastore.get_schedules({"s1"}) == [schedules[0]]
            assert datastore.get_schedules({"s2"}) == [schedules[1]]
            assert datastore.get_schedules({"s3"}) == [schedules[2]]

        for event, schedule in zip(events, schedules):
            assert event.schedule_id == schedule.id
            assert event.next_fire_time == schedule.next_fire_time

    def test_replace_schedules(
        self, datastore: DataStore, schedules: list[Schedule]
    ) -> None:
        with self.capture_events(datastore, 1, {ScheduleUpdated}) as events:
            for schedule in schedules:
                datastore.add_schedule(schedule, ConflictPolicy.exception)

            next_fire_time = schedules[2].trigger.next()
            schedule = Schedule(
                id="s3",
                task_id="foo",
                trigger=schedules[2].trigger,
                args=(),
                kwargs={},
                coalesce=CoalescePolicy.earliest,
                misfire_grace_time=None,
                tags=frozenset(),
            )
            schedule.next_fire_time = next_fire_time
            datastore.add_schedule(schedule, ConflictPolicy.replace)

            schedules = datastore.get_schedules({schedule.id})
            assert schedules[0].task_id == "foo"
            assert schedules[0].next_fire_time == next_fire_time
            assert schedules[0].args == ()
            assert schedules[0].kwargs == {}
            assert schedules[0].coalesce is CoalescePolicy.earliest
            assert schedules[0].misfire_grace_time is None
            assert schedules[0].tags == frozenset()

        received_event = events.pop(0)
        assert received_event.schedule_id == "s3"
        assert received_event.next_fire_time == datetime(
            2020, 9, 15, tzinfo=timezone.utc
        )
        assert not events

    def test_remove_schedules(
        self, datastore: DataStore, schedules: list[Schedule]
    ) -> None:
        with self.capture_events(datastore, 2, {ScheduleRemoved}) as events:
            for schedule in schedules:
                datastore.add_schedule(schedule, ConflictPolicy.exception)

            datastore.remove_schedules(["s1", "s2"])
            assert datastore.get_schedules() == [schedules[2]]

        received_event = events.pop(0)
        assert received_event.schedule_id == "s1"

        received_event = events.pop(0)
        assert received_event.schedule_id == "s2"

        assert not events

    @pytest.mark.freeze_time(datetime(2020, 9, 14, tzinfo=timezone.utc))
    def test_acquire_release_schedules(
        self, datastore: DataStore, schedules: list[Schedule]
    ) -> None:
        event_types = {ScheduleRemoved, ScheduleUpdated}
        with self.capture_events(datastore, 2, event_types) as events:
            for schedule in schedules:
                datastore.add_schedule(schedule, ConflictPolicy.exception)

            # The first scheduler gets the first due schedule
            schedules1 = datastore.acquire_schedules("dummy-id1", 1)
            assert len(schedules1) == 1
            assert schedules1[0].id == "s1"

            # The second scheduler gets the second due schedule
            schedules2 = datastore.acquire_schedules("dummy-id2", 1)
            assert len(schedules2) == 1
            assert schedules2[0].id == "s2"

            # The third scheduler gets nothing
            schedules3 = datastore.acquire_schedules("dummy-id3", 1)
            assert not schedules3

            # Update the schedules and check that the job store actually deletes the
            # first one and updates the second one
            schedules1[0].next_fire_time = None
            schedules2[0].next_fire_time = datetime(2020, 9, 15, tzinfo=timezone.utc)

            # Release all the schedules
            datastore.release_schedules("dummy-id1", schedules1)
            datastore.release_schedules("dummy-id2", schedules2)

            # Check that the first schedule is gone
            schedules = datastore.get_schedules()
            assert len(schedules) == 2
            assert schedules[0].id == "s2"
            assert schedules[1].id == "s3"

        # Check for the appropriate update and delete events
        received_event = events.pop(0)
        assert isinstance(received_event, ScheduleRemoved)
        assert received_event.schedule_id == "s1"

        received_event = events.pop(0)
        assert isinstance(received_event, ScheduleUpdated)
        assert received_event.schedule_id == "s2"
        assert received_event.next_fire_time == datetime(
            2020, 9, 15, tzinfo=timezone.utc
        )

        assert not events

    def test_release_schedule_two_identical_fire_times(
        self, datastore: DataStore
    ) -> None:
        """Regression test for #616."""
        for i in range(1, 3):
            trigger = DateTrigger(datetime(2020, 9, 13, tzinfo=timezone.utc))
            schedule = Schedule(id=f"s{i}", task_id="task1", trigger=trigger)
            schedule.next_fire_time = trigger.next()
            datastore.add_schedule(schedule, ConflictPolicy.exception)

        schedules = datastore.acquire_schedules("foo", 3)
        schedules[0].next_fire_time = None
        datastore.release_schedules("foo", schedules)

        remaining = datastore.get_schedules({s.id for s in schedules})
        assert len(remaining) == 1
        assert remaining[0].id == schedules[1].id

    def test_release_two_schedules_at_once(self, datastore: DataStore) -> None:
        """Regression test for #621."""
        for i in range(2):
            trigger = DateTrigger(datetime(2020, 9, 13, tzinfo=timezone.utc))
            schedule = Schedule(id=f"s{i}", task_id="task1", trigger=trigger)
            schedule.next_fire_time = trigger.next()
            datastore.add_schedule(schedule, ConflictPolicy.exception)

        schedules = datastore.acquire_schedules("foo", 3)
        datastore.release_schedules("foo", schedules)

        remaining = datastore.get_schedules({s.id for s in schedules})
        assert len(remaining) == 2

    def test_acquire_schedules_lock_timeout(
        self, datastore: DataStore, schedules: list[Schedule], freezer
    ) -> None:
        """
        Test that a scheduler can acquire schedules that were acquired by another
        scheduler but not released within the lock timeout period.

        """
        datastore.add_schedule(schedules[0], ConflictPolicy.exception)

        # First, one scheduler acquires the first available schedule
        acquired1 = datastore.acquire_schedules("dummy-id1", 1)
        assert len(acquired1) == 1
        assert acquired1[0].id == "s1"

        # Try to acquire the schedule just at the threshold (now == acquired_until).
        # This should not yield any schedules.
        freezer.tick(30)
        acquired2 = datastore.acquire_schedules("dummy-id2", 1)
        assert not acquired2

        # Right after that, the schedule should be available
        freezer.tick(1)
        acquired3 = datastore.acquire_schedules("dummy-id2", 1)
        assert len(acquired3) == 1
        assert acquired3[0].id == "s1"

    def test_acquire_multiple_workers(self, datastore: DataStore) -> None:
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        jobs = [Job(task_id="task1") for _ in range(2)]
        for job in jobs:
            datastore.add_job(job)

        # The first worker gets the first job in the queue
        jobs1 = datastore.acquire_jobs("worker1", 1)
        assert len(jobs1) == 1
        assert jobs1[0].id == jobs[0].id

        # The second worker gets the second job
        jobs2 = datastore.acquire_jobs("worker2", 1)
        assert len(jobs2) == 1
        assert jobs2[0].id == jobs[1].id

        # The third worker gets nothing
        jobs3 = datastore.acquire_jobs("worker3", 1)
        assert not jobs3

    def test_job_release_success(self, datastore: DataStore) -> None:
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        datastore.add_job(job)

        acquired = datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.success,
                return_value="foo",
            ),
        )
        result = datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.success
        assert result.exception is None
        assert result.return_value == "foo"

        # Check that the job and its result are gone
        assert not datastore.get_jobs({acquired[0].id})
        assert not datastore.get_job_result(acquired[0].id)

    def test_job_release_failure(self, datastore: DataStore) -> None:
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        datastore.add_job(job)

        acquired = datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.error,
                exception=ValueError("foo"),
            ),
        )
        result = datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.error
        assert isinstance(result.exception, ValueError)
        assert result.exception.args == ("foo",)
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not datastore.get_jobs({acquired[0].id})
        assert not datastore.get_job_result(acquired[0].id)

    def test_job_release_missed_deadline(self, datastore: DataStore):
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        datastore.add_job(job)

        acquired = datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.missed_start_deadline,
            ),
        )
        result = datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.missed_start_deadline
        assert result.exception is None
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not datastore.get_jobs({acquired[0].id})
        assert not datastore.get_job_result(acquired[0].id)

    def test_job_release_cancelled(self, datastore: DataStore) -> None:
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        datastore.add_job(job)

        acquired = datastore.acquire_jobs("worker1", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        datastore.release_job(
            "worker1",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.cancelled,
            ),
        )
        result = datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.cancelled
        assert result.exception is None
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not datastore.get_jobs({acquired[0].id})
        assert not datastore.get_job_result(acquired[0].id)

    def test_acquire_jobs_lock_timeout(
        self, datastore: DataStore, freezer: FrozenDateTimeFactory
    ) -> None:
        """
        Test that a worker can acquire jobs that were acquired by another scheduler but
        not released within the lock timeout period.

        """
        datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1")
        datastore.add_job(job)

        # First, one worker acquires the first available job
        acquired = datastore.acquire_jobs("worker1", 1)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        # Try to acquire the job just at the threshold (now == acquired_until).
        # This should not yield any jobs.
        freezer.tick(30)
        assert not datastore.acquire_jobs("worker2", 1)

        # Right after that, the job should be available
        freezer.tick(1)
        acquired = datastore.acquire_jobs("worker2", 1)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

    def test_acquire_jobs_max_number_exceeded(self, datastore: DataStore) -> None:
        datastore.add_task(
            Task(id="task1", func=asynccontextmanager, max_running_jobs=2)
        )
        jobs = [Job(task_id="task1"), Job(task_id="task1"), Job(task_id="task1")]
        for job in jobs:
            datastore.add_job(job)

        # Check that only 2 jobs are returned from acquire_jobs() even though the limit
        # wqas 3
        acquired_jobs = datastore.acquire_jobs("worker1", 3)
        assert [job.id for job in acquired_jobs] == [job.id for job in jobs[:2]]

        # Release one job, and the worker should be able to acquire the third job
        datastore.release_job(
            "worker1",
            acquired_jobs[0].task_id,
            JobResult.from_job(
                acquired_jobs[0],
                JobOutcome.success,
                return_value=None,
            ),
        )
        acquired_jobs = datastore.acquire_jobs("worker1", 3)
        assert [job.id for job in acquired_jobs] == [jobs[2].id]

    def test_add_get_task(self, datastore: DataStore) -> None:
        with pytest.raises(TaskLookupError):
            datastore.get_task("dummyid")

        datastore.add_task(Task(id="dummyid", func=asynccontextmanager))
        task = datastore.get_task("dummyid")
        assert task.id == "dummyid"
        assert task.func is asynccontextmanager


@pytest.mark.anyio
class TestAsyncDataStores:
    @asynccontextmanager
    async def capture_events(
        self,
        datastore: AsyncDataStore,
        limit: int,
        event_types: set[type[Event]] | None = None,
    ) -> AsyncGenerator[list[Event], None]:
        def listener(event: Event) -> None:
            events.append(event)
            if len(events) == limit:
                limit_event.set()
                subscription.unsubscribe()

        events: list[Event] = []
        limit_event = anyio.Event()
        subscription = datastore.events.subscribe(listener, event_types)
        yield events
        if limit:
            with anyio.fail_after(3):
                await limit_event.wait()

    @pytest.fixture
    async def event_broker(self) -> AsyncGenerator[AsyncEventBroker, Any]:
        broker = LocalAsyncEventBroker()
        await broker.start()
        yield broker
        await broker.stop()

    @pytest.fixture(
        params=[
            pytest.param(lazy_fixture("adapted_memory_store"), id="memory"),
            pytest.param(
                lazy_fixture("asyncpg_store"),
                id="asyncpg",
                marks=[pytest.mark.external_service],
            ),
            pytest.param(
                lazy_fixture("asyncmy_store"),
                id="asyncmy",
                marks=[pytest.mark.external_service],
            ),
        ]
    )
    async def raw_datastore(
        self, request: SubRequest, event_broker: AsyncEventBroker
    ) -> AsyncDataStore:
        return cast(AsyncDataStore, request.param)

    @pytest.fixture
    async def datastore(
        self, raw_datastore: AsyncDataStore, event_broker: AsyncEventBroker
    ) -> AsyncGenerator[AsyncDataStore, Any]:
        await raw_datastore.start(event_broker)
        yield raw_datastore
        await raw_datastore.stop()

    async def test_add_replace_task(self, datastore: AsyncDataStore) -> None:
        import math

        event_types = {TaskAdded, TaskUpdated}
        async with self.capture_events(datastore, 3, event_types) as events:
            await datastore.add_task(Task(id="test_task", func=print))
            await datastore.add_task(Task(id="test_task2", func=math.ceil))
            await datastore.add_task(Task(id="test_task", func=repr))

            tasks = await datastore.get_tasks()
            assert len(tasks) == 2
            assert tasks[0].id == "test_task"
            assert tasks[0].func is repr
            assert tasks[1].id == "test_task2"
            assert tasks[1].func is math.ceil

        received_event = events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_task"

        received_event = events.pop(0)
        assert isinstance(received_event, TaskAdded)
        assert received_event.task_id == "test_task2"

        received_event = events.pop(0)
        assert isinstance(received_event, TaskUpdated)
        assert received_event.task_id == "test_task"

        assert not events

    async def test_add_schedules(
        self, datastore: AsyncDataStore, schedules: list[Schedule]
    ) -> None:
        async with self.capture_events(datastore, 3, {ScheduleAdded}) as events:
            for schedule in schedules:
                await datastore.add_schedule(schedule, ConflictPolicy.exception)

            assert await datastore.get_schedules() == schedules
            assert await datastore.get_schedules({"s1", "s2", "s3"}) == schedules
            assert await datastore.get_schedules({"s1"}) == [schedules[0]]
            assert await datastore.get_schedules({"s2"}) == [schedules[1]]
            assert await datastore.get_schedules({"s3"}) == [schedules[2]]

        for event, schedule in zip(events, schedules):
            assert event.schedule_id == schedule.id
            assert event.next_fire_time == schedule.next_fire_time

    async def test_replace_schedules(
        self, datastore: AsyncDataStore, schedules: list[Schedule]
    ) -> None:
        async with self.capture_events(datastore, 1, {ScheduleUpdated}) as events:
            for schedule in schedules:
                await datastore.add_schedule(schedule, ConflictPolicy.exception)

            next_fire_time = schedules[2].trigger.next()
            schedule = Schedule(
                id="s3",
                task_id="foo",
                trigger=schedules[2].trigger,
                args=(),
                kwargs={},
                coalesce=CoalescePolicy.earliest,
                misfire_grace_time=None,
                tags=frozenset(),
            )
            schedule.next_fire_time = next_fire_time
            await datastore.add_schedule(schedule, ConflictPolicy.replace)

            schedules = await datastore.get_schedules({schedule.id})
            assert schedules[0].task_id == "foo"
            assert schedules[0].next_fire_time == next_fire_time
            assert schedules[0].args == ()
            assert schedules[0].kwargs == {}
            assert schedules[0].coalesce is CoalescePolicy.earliest
            assert schedules[0].misfire_grace_time is None
            assert schedules[0].tags == frozenset()

        received_event = events.pop(0)
        assert received_event.schedule_id == "s3"
        assert received_event.next_fire_time == datetime(
            2020, 9, 15, tzinfo=timezone.utc
        )
        assert not events

    async def test_remove_schedules(
        self, datastore: AsyncDataStore, schedules: list[Schedule]
    ) -> None:
        async with self.capture_events(datastore, 2, {ScheduleRemoved}) as events:
            for schedule in schedules:
                await datastore.add_schedule(schedule, ConflictPolicy.exception)

            await datastore.remove_schedules(["s1", "s2"])
            assert await datastore.get_schedules() == [schedules[2]]

        received_event = events.pop(0)
        assert received_event.schedule_id == "s1"

        received_event = events.pop(0)
        assert received_event.schedule_id == "s2"

        assert not events

    @pytest.mark.freeze_time(datetime(2020, 9, 14, tzinfo=timezone.utc))
    async def test_acquire_release_schedules(
        self, datastore: AsyncDataStore, schedules: list[Schedule]
    ) -> None:
        event_types = {ScheduleRemoved, ScheduleUpdated}
        async with self.capture_events(datastore, 2, event_types) as events:
            for schedule in schedules:
                await datastore.add_schedule(schedule, ConflictPolicy.exception)

            # The first scheduler gets the first due schedule
            schedules1 = await datastore.acquire_schedules("dummy-id1", 1)
            assert len(schedules1) == 1
            assert schedules1[0].id == "s1"

            # The second scheduler gets the second due schedule
            schedules2 = await datastore.acquire_schedules("dummy-id2", 1)
            assert len(schedules2) == 1
            assert schedules2[0].id == "s2"

            # The third scheduler gets nothing
            schedules3 = await datastore.acquire_schedules("dummy-id3", 1)
            assert not schedules3

            # Update the schedules and check that the job store actually deletes the
            # first one and updates the second one
            schedules1[0].next_fire_time = None
            schedules2[0].next_fire_time = datetime(2020, 9, 15, tzinfo=timezone.utc)

            # Release all the schedules
            await datastore.release_schedules("dummy-id1", schedules1)
            await datastore.release_schedules("dummy-id2", schedules2)

            # Check that the first schedule is gone
            schedules = await datastore.get_schedules()
            assert len(schedules) == 2
            assert schedules[0].id == "s2"
            assert schedules[1].id == "s3"

        # Check for the appropriate update and delete events
        received_event = events.pop(0)
        assert isinstance(received_event, ScheduleRemoved)
        assert received_event.schedule_id == "s1"

        received_event = events.pop(0)
        assert isinstance(received_event, ScheduleUpdated)
        assert received_event.schedule_id == "s2"
        assert received_event.next_fire_time == datetime(
            2020, 9, 15, tzinfo=timezone.utc
        )

        assert not events

    async def test_release_schedule_two_identical_fire_times(
        self, datastore: AsyncDataStore
    ) -> None:
        """Regression test for #616."""
        for i in range(1, 3):
            trigger = DateTrigger(datetime(2020, 9, 13, tzinfo=timezone.utc))
            schedule = Schedule(id=f"s{i}", task_id="task1", trigger=trigger)
            schedule.next_fire_time = trigger.next()
            await datastore.add_schedule(schedule, ConflictPolicy.exception)

        schedules = await datastore.acquire_schedules("foo", 3)
        schedules[0].next_fire_time = None
        await datastore.release_schedules("foo", schedules)

        remaining = await datastore.get_schedules({s.id for s in schedules})
        assert len(remaining) == 1
        assert remaining[0].id == schedules[1].id

    async def test_release_two_schedules_at_once(
        self, datastore: AsyncDataStore
    ) -> None:
        """Regression test for #621."""
        for i in range(2):
            trigger = DateTrigger(datetime(2020, 9, 13, tzinfo=timezone.utc))
            schedule = Schedule(id=f"s{i}", task_id="task1", trigger=trigger)
            schedule.next_fire_time = trigger.next()
            await datastore.add_schedule(schedule, ConflictPolicy.exception)

        schedules = await datastore.acquire_schedules("foo", 3)
        await datastore.release_schedules("foo", schedules)

        remaining = await datastore.get_schedules({s.id for s in schedules})
        assert len(remaining) == 2

    async def test_acquire_schedules_lock_timeout(
        self, datastore: AsyncDataStore, schedules: list[Schedule], freezer
    ) -> None:
        """
        Test that a scheduler can acquire schedules that were acquired by another
        scheduler but not released within the lock timeout period.

        """
        await datastore.add_schedule(schedules[0], ConflictPolicy.exception)

        # First, one scheduler acquires the first available schedule
        acquired1 = await datastore.acquire_schedules("dummy-id1", 1)
        assert len(acquired1) == 1
        assert acquired1[0].id == "s1"

        # Try to acquire the schedule just at the threshold (now == acquired_until).
        # This should not yield any schedules.
        freezer.tick(30)
        acquired2 = await datastore.acquire_schedules("dummy-id2", 1)
        assert not acquired2

        # Right after that, the schedule should be available
        freezer.tick(1)
        acquired3 = await datastore.acquire_schedules("dummy-id2", 1)
        assert len(acquired3) == 1
        assert acquired3[0].id == "s1"

    async def test_acquire_multiple_workers(self, datastore: AsyncDataStore) -> None:
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        jobs = [Job(task_id="task1") for _ in range(2)]
        for job in jobs:
            await datastore.add_job(job)

        # The first worker gets the first job in the queue
        jobs1 = await datastore.acquire_jobs("worker1", 1)
        assert len(jobs1) == 1
        assert jobs1[0].id == jobs[0].id

        # The second worker gets the second job
        jobs2 = await datastore.acquire_jobs("worker2", 1)
        assert len(jobs2) == 1
        assert jobs2[0].id == jobs[1].id

        # The third worker gets nothing
        jobs3 = await datastore.acquire_jobs("worker3", 1)
        assert not jobs3

    async def test_job_release_success(self, datastore: AsyncDataStore) -> None:
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        await datastore.add_job(job)

        acquired = await datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        await datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.success,
                return_value="foo",
            ),
        )
        result = await datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.success
        assert result.exception is None
        assert result.return_value == "foo"

        # Check that the job and its result are gone
        assert not await datastore.get_jobs({acquired[0].id})
        assert not await datastore.get_job_result(acquired[0].id)

    async def test_job_release_failure(self, datastore: AsyncDataStore) -> None:
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        await datastore.add_job(job)

        acquired = await datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        await datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.error,
                exception=ValueError("foo"),
            ),
        )
        result = await datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.error
        assert isinstance(result.exception, ValueError)
        assert result.exception.args == ("foo",)
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not await datastore.get_jobs({acquired[0].id})
        assert not await datastore.get_job_result(acquired[0].id)

    async def test_job_release_missed_deadline(self, datastore: AsyncDataStore):
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        await datastore.add_job(job)

        acquired = await datastore.acquire_jobs("worker_id", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        await datastore.release_job(
            "worker_id",
            acquired[0].task_id,
            JobResult.from_job(
                acquired[0],
                JobOutcome.missed_start_deadline,
            ),
        )
        result = await datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.missed_start_deadline
        assert result.exception is None
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not await datastore.get_jobs({acquired[0].id})
        assert not await datastore.get_job_result(acquired[0].id)

    async def test_job_release_cancelled(self, datastore: AsyncDataStore) -> None:
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        await datastore.add_job(job)

        acquired = await datastore.acquire_jobs("worker1", 2)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        await datastore.release_job(
            "worker1",
            acquired[0].task_id,
            JobResult.from_job(acquired[0], JobOutcome.cancelled),
        )
        result = await datastore.get_job_result(acquired[0].id)
        assert result.outcome is JobOutcome.cancelled
        assert result.exception is None
        assert result.return_value is None

        # Check that the job and its result are gone
        assert not await datastore.get_jobs({acquired[0].id})
        assert not await datastore.get_job_result(acquired[0].id)

    async def test_acquire_jobs_lock_timeout(
        self, datastore: AsyncDataStore, freezer: FrozenDateTimeFactory
    ) -> None:
        """
        Test that a worker can acquire jobs that were acquired by another scheduler but
        not released within the lock timeout period.

        """
        await datastore.add_task(Task(id="task1", func=asynccontextmanager))
        job = Job(task_id="task1", result_expiration_time=timedelta(minutes=1))
        await datastore.add_job(job)

        # First, one worker acquires the first available job
        acquired = await datastore.acquire_jobs("worker1", 1)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

        # Try to acquire the job just at the threshold (now == acquired_until).
        # This should not yield any jobs.
        freezer.tick(30)
        assert not await datastore.acquire_jobs("worker2", 1)

        # Right after that, the job should be available
        freezer.tick(1)
        acquired = await datastore.acquire_jobs("worker2", 1)
        assert len(acquired) == 1
        assert acquired[0].id == job.id

    async def test_acquire_jobs_max_number_exceeded(
        self, datastore: AsyncDataStore
    ) -> None:
        await datastore.add_task(
            Task(id="task1", func=asynccontextmanager, max_running_jobs=2)
        )
        jobs = [Job(task_id="task1"), Job(task_id="task1"), Job(task_id="task1")]
        for job in jobs:
            await datastore.add_job(job)

        # Check that only 2 jobs are returned from acquire_jobs() even though the limit
        # wqas 3
        acquired_jobs = await datastore.acquire_jobs("worker1", 3)
        assert [job.id for job in acquired_jobs] == [job.id for job in jobs[:2]]

        # Release one job, and the worker should be able to acquire the third job
        await datastore.release_job(
            "worker1",
            acquired_jobs[0].task_id,
            JobResult.from_job(
                acquired_jobs[0],
                JobOutcome.success,
                return_value=None,
            ),
        )
        acquired_jobs = await datastore.acquire_jobs("worker1", 3)
        assert [job.id for job in acquired_jobs] == [jobs[2].id]

    async def test_add_get_task(self, datastore: DataStore) -> None:
        with pytest.raises(TaskLookupError):
            await datastore.get_task("dummyid")

        await datastore.add_task(Task(id="dummyid", func=asynccontextmanager))
        task = await datastore.get_task("dummyid")
        assert task.id == "dummyid"
        assert task.func is asynccontextmanager

    async def test_cancel_start(
        self, raw_datastore: AsyncDataStore, event_broker: AsyncEventBroker
    ) -> None:
        with CancelScope() as scope:
            scope.cancel()
            await raw_datastore.start(event_broker)
            await raw_datastore.stop()

    async def test_cancel_stop(
        self, raw_datastore: AsyncDataStore, event_broker: AsyncEventBroker
    ) -> None:
        with CancelScope() as scope:
            await raw_datastore.start(event_broker)
            scope.cancel()
            await raw_datastore.stop()



from __future__ import annotations

from datetime import datetime, timedelta

import pytest

from apscheduler.triggers.interval import IntervalTrigger


def test_bad_interval():
    exc = pytest.raises(ValueError, IntervalTrigger)
    exc.match("The time interval must be positive")


def test_bad_end_time(timezone):
    start_time = datetime(2020, 5, 16, tzinfo=timezone)
    end_time = datetime(2020, 5, 15, tzinfo=timezone)
    exc = pytest.raises(
        ValueError, IntervalTrigger, seconds=1, start_time=start_time, end_time=end_time
    )
    exc.match("end_time cannot be earlier than start_time")


def test_end_time(timezone, serializer):
    start_time = datetime(2020, 5, 16, 19, 32, 44, 649521, tzinfo=timezone)
    end_time = datetime(2020, 5, 16, 22, 33, 1, tzinfo=timezone)
    interval = timedelta(hours=1, seconds=6)
    trigger = IntervalTrigger(
        start_time=start_time, end_time=end_time, hours=1, seconds=6
    )
    if serializer:
        trigger = serializer.deserialize(serializer.serialize(trigger))

    assert trigger.next() == start_time
    assert trigger.next() == start_time + interval
    assert trigger.next() == start_time + interval * 2
    assert trigger.next() is None


def test_repr(timezone, serializer):
    start_time = datetime(2020, 5, 15, 12, 55, 32, 954032, tzinfo=timezone)
    end_time = datetime(2020, 6, 4, 16, 18, 49, 306942, tzinfo=timezone)
    trigger = IntervalTrigger(
        weeks=1,
        days=2,
        hours=3,
        minutes=4,
        seconds=5,
        microseconds=123525,
        start_time=start_time,
        end_time=end_time,
    )
    if serializer:
        trigger = serializer.deserialize(serializer.serialize(trigger))

    assert repr(trigger) == (
        "IntervalTrigger(weeks=1, days=2, hours=3, minutes=4, seconds=5, "
        "microseconds=123525, start_time='2020-05-15 12:55:32.954032+02:00', "
        "end_time='2020-06-04 16:18:49.306942+02:00')"
    )



from __future__ import annotations

from asyncio import iscoroutine
from typing import Any, Callable

import attrs
from anyio import create_task_group
from anyio.abc import TaskGroup

from .._events import Event
from ..abc import AsyncEventBroker
from .base import BaseEventBroker


@attrs.define(eq=False)
class LocalAsyncEventBroker(AsyncEventBroker, BaseEventBroker):
    """
    Asynchronous, local event broker.

    This event broker only broadcasts within the process it runs in, and is therefore
    not suitable for multi-node or multiprocess use cases.

    Does not serialize events.
    """

    _task_group: TaskGroup = attrs.field(init=False)

    async def start(self) -> None:
        self._task_group = create_task_group()
        await self._task_group.__aenter__()

    async def stop(self, *, force: bool = False) -> None:
        await self._task_group.__aexit__(None, None, None)
        del self._task_group

    async def publish(self, event: Event) -> None:
        await self.publish_local(event)

    async def publish_local(self, event: Event) -> None:
        event_type = type(event)
        one_shot_tokens: list[object] = []
        for _token, subscription in self._subscriptions.items():
            if (
                subscription.event_types is None
                or event_type in subscription.event_types
            ):
                self._task_group.start_soon(
                    self._deliver_event, subscription.callback, event
                )
                if subscription.one_shot:
                    one_shot_tokens.append(subscription.token)

        for token in one_shot_tokens:
            super().unsubscribe(token)

    async def _deliver_event(self, func: Callable[[Event], Any], event: Event) -> None:
        try:
            retval = func(event)
            if iscoroutine(retval):
                await retval
        except BaseException:
            self._logger.exception(
                "Error delivering %s event", event.__class__.__name__
            )



from __future__ import annotations

from concurrent.futures import Future
from typing import Any

import attrs
from paho.mqtt.client import Client, MQTTMessage
from paho.mqtt.properties import Properties
from paho.mqtt.reasoncodes import ReasonCodes

from .._events import Event
from ..abc import Serializer
from ..serializers.json import JSONSerializer
from .base import DistributedEventBrokerMixin
from .local import LocalEventBroker


@attrs.define(eq=False)
class MQTTEventBroker(LocalEventBroker, DistributedEventBrokerMixin):
    """
    An event broker that uses an MQTT (v3.1 or v5) broker to broadcast events.

    Requires the paho-mqtt_ library to be installed.

    .. _paho-mqtt: https://pypi.org/project/paho-mqtt/

    :param client: a paho-mqtt client
    :param serializer: the serializer used to (de)serialize events for transport
    :param host: host name or IP address to connect to
    :param port: TCP port number to connect to
    :param topic: topic on which to send the messages
    :param subscribe_qos: MQTT QoS to use for subscribing messages
    :param publish_qos: MQTT QoS to use for publishing messages
    """

    client: Client
    serializer: Serializer = attrs.field(factory=JSONSerializer)
    host: str = attrs.field(kw_only=True, default="localhost")
    port: int = attrs.field(kw_only=True, default=1883)
    topic: str = attrs.field(kw_only=True, default="apscheduler")
    subscribe_qos: int = attrs.field(kw_only=True, default=0)
    publish_qos: int = attrs.field(kw_only=True, default=0)
    _ready_future: Future[None] = attrs.field(init=False)

    def start(self) -> None:
        super().start()
        self._ready_future = Future()
        self.client.enable_logger(self._logger)
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_subscribe = self._on_subscribe
        self.client.connect(self.host, self.port)
        self.client.loop_start()
        self._ready_future.result(10)

    def stop(self, *, force: bool = False) -> None:
        self.client.disconnect()
        self.client.loop_stop(force=force)
        super().stop()

    def _on_connect(
        self,
        client: Client,
        userdata: Any,
        flags: dict[str, Any],
        rc: ReasonCodes | int,
        properties: Properties | None = None,
    ) -> None:
        try:
            client.subscribe(self.topic, qos=self.subscribe_qos)
        except Exception as exc:
            self._ready_future.set_exception(exc)
            raise

    def _on_subscribe(
        self, client: Client, userdata: Any, mid, granted_qos: list[int]
    ) -> None:
        self._ready_future.set_result(None)

    def _on_message(self, client: Client, userdata: Any, msg: MQTTMessage) -> None:
        event = self.reconstitute_event(msg.payload)
        if event is not None:
            self.publish_local(event)

    def publish(self, event: Event) -> None:
        notification = self.generate_notification(event)
        self.client.publish(self.topic, notification, qos=self.publish_qos)



from __future__ import annotations

from base64 import b64decode, b64encode
from logging import Logger, getLogger
from typing import Any, Callable, Iterable

import attrs

from .. import _events
from .._events import Event
from .._exceptions import DeserializationError
from ..abc import EventSource, Serializer, Subscription


@attrs.define(eq=False, frozen=True)
class LocalSubscription(Subscription):
    callback: Callable[[Event], Any]
    event_types: set[type[Event]] | None
    one_shot: bool
    token: object
    _source: BaseEventBroker

    def unsubscribe(self) -> None:
        self._source.unsubscribe(self.token)


@attrs.define(eq=False)
class BaseEventBroker(EventSource):
    _logger: Logger = attrs.field(init=False)
    _subscriptions: dict[object, LocalSubscription] = attrs.field(
        init=False, factory=dict
    )

    def __attrs_post_init__(self) -> None:
        self._logger = getLogger(self.__class__.__module__)

    def subscribe(
        self,
        callback: Callable[[Event], Any],
        event_types: Iterable[type[Event]] | None = None,
        *,
        one_shot: bool = False,
    ) -> Subscription:
        types = set(event_types) if event_types else None
        token = object()
        subscription = LocalSubscription(callback, types, one_shot, token, self)
        self._subscriptions[token] = subscription
        return subscription

    def unsubscribe(self, token: object) -> None:
        self._subscriptions.pop(token, None)


class DistributedEventBrokerMixin:
    serializer: Serializer
    _logger: Logger

    def generate_notification(self, event: Event) -> bytes:
        serialized = self.serializer.serialize(event.marshal(self.serializer))
        return event.__class__.__name__.encode("ascii") + b" " + serialized

    def generate_notification_str(self, event: Event) -> str:
        serialized = self.serializer.serialize(event.marshal(self.serializer))
        return event.__class__.__name__ + " " + b64encode(serialized).decode("ascii")

    def _reconstitute_event(self, event_type: str, serialized: bytes) -> Event | None:
        try:
            kwargs = self.serializer.deserialize(serialized)
        except DeserializationError:
            self._logger.exception(
                "Failed to deserialize an event of type %s",
                event_type,
                extra={"serialized": serialized},
            )
            return None

        try:
            event_class = getattr(_events, event_type)
        except AttributeError:
            self._logger.error(
                "Receive notification for a nonexistent event type: %s",
                event_type,
                extra={"serialized": serialized},
            )
            return None

        try:
            return event_class.unmarshal(self.serializer, kwargs)
        except Exception:
            self._logger.exception("Error reconstituting event of type %s", event_type)
            return None

    def reconstitute_event(self, payload: bytes) -> Event | None:
        try:
            event_type_bytes, serialized = payload.split(b" ", 1)
        except ValueError:
            self._logger.error(
                "Received malformatted notification", extra={"payload": payload}
            )
            return None

        event_type = event_type_bytes.decode("ascii", errors="replace")
        return self._reconstitute_event(event_type, serialized)

    def reconstitute_event_str(self, payload: str) -> Event | None:
        try:
            event_type, b64_serialized = payload.split(" ", 1)
        except ValueError:
            self._logger.error(
                "Received malformatted notification", extra={"payload": payload}
            )
            return None

        return self._reconstitute_event(event_type, b64decode(b64_serialized))
